////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "voxelintegrator.h"
#include "geometry/engine.h"
#include "viewer/d3layer.h"
#include "viewer/viewer.h"
#include "math/randsampler.h"
#include "rendermaterial.h"

#include <iostream>
#include <fstream>
#include <map>

using namespace std;

CVoxelIntegrator::CVoxelIntegrator() : CBaseIntegrator(), m_MaxVRayDepth(1000000), m_RenderPass(256), m_TransmissionThreshold(0.10, 0.10, 0.10), m_OpticalDensity(0.2),
																			 m_Shine(0.5), m_RayMarchDelta(-1), m_AOCRayMarchDelta(-1), m_bShadows( true ), m_AmbientNormalization(75.0),
																			 m_AOCMinIntensity( 0. ), m_AOCMaxIntensity( 1.0 ), m_AOCMaxDist( 10 ), m_NrOfAOCSamples( 20 ), m_TransferFuncIndex(0), m_bMIPProjection(false),
																			 m_MinMIPThreshold( -1e99 )
{
}

CVoxelIntegrator::~CVoxelIntegrator( void )
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::InitFromConfig()

// [6/29/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert( m_pRenderer );	// ensure the the base-class Init() has been called before

	m_OpticalDensity		= m_pRenderer->GetConfig().GetValue<double>( "optical_density", 0.2 );
	m_TransferFuncIndex = m_pRenderer->GetConfig().GetValue<int>( "transfer_func", 0 );
	m_Shine							= m_pRenderer->GetConfig().GetValue<double>( "volume_specular", 0.5 );
	m_RayMarchDelta			= m_pRenderer->GetConfig().GetValue<double>( "volume_ray-delta", -1 );
	m_AOCMinIntensity		= m_pRenderer->GetConfig().GetValue<double>( "volume_aoc-min", 0.0 );
	m_AOCMaxIntensity		= m_pRenderer->GetConfig().GetValue<double>( "volume_aoc-max", 1.0 );
	m_bMIPProjection		= m_pRenderer->GetConfig().GetValue<bool>( "volume_MIP", 0 );
	m_MinMIPThreshold		= m_pRenderer->GetConfig().GetValue<double>( "volume_min-MIP", -1e99 );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CVoxelIntegrator::Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// main volume/voxel integrator routine returning the luminance for the passed ray & rendering pass
// [6/9/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Point	 Hit;
	Color	 ShadingColor, TransmissionCol;
	Ray		 lightRay;
	double randNr[2];
	double AmbiScaling = 1.0 / m_AmbientNormalization;

// 	ExportColorHisto( "color-table.txt" );
// 	exit(0);

	m_RenderPass = RenderPass;

	assert( m_pVoxelLayer );		// this integrator should never be called without a corresponding voxel object layer!

	// on end of ray, error or blackbody
	if( !IdentifyVoxelRegion( pRay ) ) {
		pRay->depth = 0;
		if( m_pVisLayer->nextIntersection( pEngine, pRay ) )
			return m_pVisLayer->shade( pEngine, pRay, RenderPass );
		else
			return Color::Black;
	}

	// Get region/voxel color
	GetVoxelColor( pRay, ShadingColor );

	// Calculate illumination
	Color VoxelTransparency = RayToTransparency( pRay );

	// if requested we do MIP rendering instead of standard transmission composition.
	// instead of doing a full shading + recursive transmission compositing we directly scan until the end
	// of the voxel geometry identifying the voxel with the maximum intensity and do the shading on that one only.
	if( m_bMIPProjection ) {
		ShadingColor = m_pVoxelLayer->color( pRay->voxelreg );	// initialize MIP with first raw data color
		Li_MIP( pEngine, pRay, ShadingColor, VoxelTransparency );	// evaluate MIP along the ray

		// check if we are below the MIP threshold and have to consider objects behind the voxel geometry
		if( m_pVoxelLayer->color2value( ShadingColor.color32() ) < m_MinMIPThreshold ) {
			return Color::Black;
		}

		AmbiScaling = 4.0 / 255.0;								// depending on the render quality we need different ambient boosts
	}

	// If we are in the start of the plane do not apply any shadow
	// if( pRay->start() ) return ShadingColor;

	// All light rays should start from outside the object
	// Move a bit backwards
	// Should be bigger than what is inside distance
	Hit = pRay->hit( -SMALL3D3 );

	// to emphasize different materials in the shading the voxel's gradient is used instead of
	// the standard ray's normal vector only
	Vector Normal = GetShadingNormal( pRay );


	// we only go through the lighting calc if the voxel is not completely transparent
	// and if we are not doing MIP, which does not know about transparency and would otherwise
	// result in strange lighting/shading artifacts
	if( !m_bMIPProjection && VoxelTransparency != Color::White ) {
		double totalIntensity = 0.0, curIntensity = 0.0;
		Vector LightDir, curLPos;

		for( int i = 0; i < Min( pRay->nlights, m_pGeometry->nlights ); i++ ) {
			const Light& L = m_pGeometry->light[Light0 + i];	// Geometry Light properties
			const VLight& VL = m_pVisLayer->light[i];					// Viewer light coordinate & direction
			double  dist, power, penumbraScaling = 0.0;

			// CT: we only do volumetric lights during the beauty pass
			const int nsamples = L.nsamples;
			//assert(InRange(0,nsamples,100));
			curIntensity = 0.0;

			// volumetric lights
			for( int lightSample = 0; lightSample < nsamples; ++lightSample ) {
				switch( L.type ) {
					case Light::Type::Sun:
						LightDir = -VL.dir;
						dist = INFINITE;
						power = L.power;
						if( m_bShadows && L.shadow && pRay->shadow ) {
							double dot = m_pVisLayer->view().matrix( 0, 2 ) * LightDir.x +
								m_pVisLayer->view().matrix( 1, 2 ) * LightDir.y +
								m_pVisLayer->view().matrix( 2, 2 ) * LightDir.z;
							if( dot > SMALL3D3 && m_pVisLayer->view().projection == ViewPort::Projection::Orthographic ) {
								dist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
												 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
												 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
									dot;
								if( dist < SMALL3D3 ) continue;
							}
						}
						break;

					case Light::Type::Omni:
					case Light::Type::Spot:
						// move the light position for volumetric light sources
						curLPos = VL.pos;
						if( lightSample > 0 ) {
							randNr[0] = CBaseSampler::VanDerCorput( lightSample, pEngine->random.integer() );
							randNr[1] = CBaseSampler::Sobol2( lightSample, pEngine->random.integer() );
							LightDir = (VL.pos - Hit);
							LightDir.normalize();
							curLPos += CBaseSampler::uniformSampleDisk( LightDir, L.radius, randNr[0], randNr[1] );
						}
						LightDir = curLPos - Hit;
						dist = LightDir.normalize();

						if( dist < SMALL3D3 ) continue;
						if( dist > L.distance ) continue;
						if( L.falloff != Light::FallOff::Constant ) {
							// FIXME could be converted to switch ()
							if( L.falloff == Light::FallOff::Inverse )
								power = L.power * 100.0 / dist;
							else	// must be Light::FallOff::Inverse2
								power = L.power * 10000.0 / Sq( dist );
							if( power < 0.001 ) continue;
						}
						else
							power = L.power;

						if( m_bShadows && L.shadow && pRay->shadow ) {
							double dot = m_pVisLayer->view().matrix( 0, 2 ) * LightDir.x +
								m_pVisLayer->view().matrix( 1, 2 ) * LightDir.y +
								m_pVisLayer->view().matrix( 2, 2 ) * LightDir.z;
							if( dot > SMALL3D3 ) {
								double pdist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
																 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
																 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
									dot;
								dist = Min( dist, pdist );
								if( dist < SMALL3D3 ) continue;
							}
						}
						power /= (double)nsamples;   // normalize the light power w.r.t. light samples
						break;

					default:
						continue;
				}

				// set up light ray & find intersection with light
				lightRay.use_clip = pRay->use_clip;
				RaySegment new_segment( Hit, LightDir, pRay->prevZone() );
				new_segment.tmin = SMALL3D3 * pRay->segment().tmin;
				new_segment.tmax = dist;
				lightRay.push( new_segment );

				double dot = LightDir * Normal;
				if( dot > 0.00001 ) {
					if( m_bShadows && L.shadow && pRay->shadow ) {

						int oldBodyCheckId = pEngine->incBodyCheckId();

						// integrate transparency along the shadow ray
						Color IntTransparency( 1.0, 1.0, 1.0 );
						IntegrateTransparency( &lightRay, IntTransparency );
						penumbraScaling += IntTransparency.luminance() / (double)nsamples;

						pEngine->bodyCheckId( oldBodyCheckId );
					}
				}
				else // no shadowing requested
					penumbraScaling = 1.0;

				// standard diffuse shading
				curIntensity += power * dot;

				///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
				// add shininess to shader
				if( m_Shine > 0.0 ) {
					Vector Half = -pRay->dir() + LightDir;
					Half.normalize();
					double Spec = Half * Normal;

					if( Spec > 0.0 )
						curIntensity += power * pow( Spec, m_Shine );
				}

				// bail out early after 1 sample in case of sunlight
				if( L.type == Light::Type::Sun && lightSample == 0 )
					break;
			} // end of sampling for one light source

			totalIntensity += curIntensity * penumbraScaling;

		} // end of light loop

		// standard shading
		ShadingColor *= (totalIntensity + (float)m_pVisLayer->ambient * AmbiScaling);							// NOTE: If we use the gradients as shading normals we need a bit more
																																															//			 ambient contribution
	} // end of shading calculation

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// local ambient occlusion for opaque voxels
	if( !m_bMIPProjection && VoxelTransparency < m_TransmissionThreshold &&
			m_RenderQuality >= 4 ) {
		double AOCShading = CalcAmbientOcclusion( pEngine, pRay );
		ShadingColor *= AOCShading;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// standard transmisison composite rendering
	//
	//
	if( !m_bMIPProjection && pRay->depth < m_MaxVRayDepth && VoxelTransparency > m_TransmissionThreshold ) {
		TransmissionCol = Li_Transmission( pEngine, pRay, RenderPass, Light0 );
		ShadingColor.blend( TransmissionCol, Color::White - VoxelTransparency );
	}

	ShadingColor.clamp();

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::Li_MIP( GeometryEngine* pEngine, Ray* pRay, Color& CurVoxelColor, Color& CurVoxelTransparency )

// [10/23/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color NextVoxelColor;

	if( m_RayMarchDelta < 0 )
		m_RayMarchDelta = std::min( m_pVoxelLayer->dx(), std::min( m_pVoxelLayer->dy(), m_pVoxelLayer->dz() ) ) * 0.5;

	// march forward
	pRay->segment().tmin += m_RayMarchDelta;
	pRay->voxelreg = -1;
	pRay->incDepth();

		// on end of ray, error or blackbody
	if( !IdentifyVoxelRegion( pRay ) ) {
		pRay->depth = 0;
		return;
	}

	NextVoxelColor = m_pVoxelLayer->color( pRay->voxelreg );	// use raw data instead of interpolated!

	// try to find the maximum intensity when walking along the ray
	if( m_pVoxelLayer->color2value( CurVoxelColor.color32() ) < m_pVoxelLayer->color2value( NextVoxelColor.color32() ) &&
	m_pVoxelLayer->color2value( NextVoxelColor.color32() ) >= m_MinMIPThreshold ) {
		CurVoxelColor = NextVoxelColor;
		CurVoxelTransparency = Color::Black;
	}

	Li_MIP( pEngine, pRay, CurVoxelColor, CurVoxelTransparency);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CVoxelIntegrator::Li_Transmission( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [6/9/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color						TransmissionColor, TransparencyColor( 1.0, 1.0, 1.0 );

	//
	// march ray
	// TODO: We could think about taking larger steps until a different material is found
	// TODO: This should be set only once before the rendering starts
	//
	if( m_RayMarchDelta <= 0 ) {
		m_RayMarchDelta			= std::min( m_pVoxelLayer->dx(), std::min( m_pVoxelLayer->dy(), m_pVoxelLayer->dz() ) ) * 0.5;
		m_AOCRayMarchDelta	= std::max( m_pVoxelLayer->dx(), std::min( m_pVoxelLayer->dy(), m_pVoxelLayer->dz() ) );
	}

	pRay->segment().tmin += m_RayMarchDelta;	// march forward
	pRay->voxelreg = -1;
	pRay->incDepth();
	TransmissionColor = Li( pEngine, pRay, RenderPass, Light0 );

	//
	// apply Beer's law
	// for optical absorption, which makes quite a difference in terms of image quality
	if( pRay->voxelreg != -1) {
		double MatDensity = -m_OpticalDensity * (m_RayMarchDelta);
		Color Absorbance( m_pVoxelLayer->color( pRay->voxelreg ) );
		Absorbance *= MatDensity;
		TransparencyColor.set( expf( Absorbance.red() ), expf( Absorbance.green() ), expf( Absorbance.blue() ) );
	}

	return TransmissionColor * TransparencyColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color	CVoxelIntegrator::RayToTransparency( Ray* pRay )

// derive the transparency from the hit ray
// [6/9/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color	Transparency;

	Color32 CurColor = m_pVoxelLayer->color( pRay->voxelreg );

	// TODO: define more sophisticated color to transparency function based on gradients
	//			 In the absence of a GUI we hard-code two different functions that can be selected
	//			 via the farm.ini file

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// TEST transparency for Differ palette (inverted!)
	// Do NOT remove and change this
// 	Transparency = 0.5;
// 	if( CurColor.red() > 240 && CurColor.green() > 220 && CurColor.blue() > 220 )
// 		Transparency = 1.0;
// 	else if( CurColor.red() < 50 && CurColor.green() < 50 && CurColor.blue() > 220 )
// 		Transparency = 0.0;
// 	else if( CurColor.blue() > 220 )
// 		Transparency = 0.0;
// 	else if( CurColor.red() > 250 )
// 		Transparency = 0.98;
// 	else
// 		Transparency = CurColor.red() / 256.0;
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	if( m_TransferFuncIndex == 0 ) {
		Transparency.set( 0.5, 0.5, 0.5 );
		if( CurColor.red() > 240 && CurColor.green() > 220 && CurColor.blue() > 220 )
			Transparency.set(1.0, 1.0, 1.0);
		else if( CurColor.red() < 50 && CurColor.green() < 50 && CurColor.blue() > 220 )
			Transparency.set( 0.0, 0.0, 0.0 );
		else if( CurColor.blue() > 220 )
			Transparency.set( 0.0, 0.0, 0.0 );
		else if( CurColor.red() > 250 )
			Transparency.set( 0.98, 0.98, 0.98 );
		else
			Transparency.set( CurColor.red() / 256.0, CurColor.red() / 256.0, CurColor.red() / 256.0);
	}
	else if( m_TransferFuncIndex == 1 ) {
		Transparency.set( 0.98, 0.98, 0.98 );
		if( CurColor.red() < 20 && CurColor.green() < 20 && CurColor.blue() < 20 )
			Transparency.set( 0.0, 0.0, 0.0 );
		else if( CurColor.red() < 50 && CurColor.green() < 50 && CurColor.blue() < 50 )
			Transparency.set( 0.2, 0.2, 0.2 );
	}
	else {		// try a multi-dimensional transfer function
		Transparency.set( 0.98, 0.85, 0.85 );
		if( CurColor.red() < 20 && CurColor.green() < 20 && CurColor.blue() < 20 )
			Transparency.set( 0.0, 0.0, 0.0 );
		else if( CurColor.red() < 50 && CurColor.green() < 50 && CurColor.blue() < 50 )
			Transparency.set( 0.4, 0.3, 0.2 );
	}

	return Transparency;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CVoxelIntegrator::CalcNormalizedVoxelGradient( Ray* pRay )

// determine the gradient vector for the current voxel region based on it's color.
// NOTE: ideally we would use the data value
// [6/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Vector GradientVec, TempGradVec;
	int		 x = -1, y = -1, z = -1;

	// FIXME: This needs to be made thread safe!
	//if( GetNormalizedGradientFromCache( pRay->voxelreg, GradientVec ) )
	//	return GradientVec;

	// translate region nr into indices
	if( !m_pVoxelLayer->index2ijk( pRay->voxelreg, x, y, z ) )
		return GradientVec;

	//
	CalcVoxelGradient( x, y, z, GradientVec );
	GradientVec.normalize();

	// add to cache
	// AddGradientToCache( pRay->voxelreg, GradientVec );

	return GradientVec;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::CalcVoxelGradient( int i, int j, int k, Vector& GradientVec )

// calculate the gradient of voxel identified by i, j, k by differentiation
// NOTE: This is uncached!!
//
// [6/15/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	GradientVec.x = (m_pVoxelLayer->color( i + 1, j    , k     ) - m_pVoxelLayer->color( i - 1, j    , k     )) / 255.0 / (2.0 * m_pVoxelLayer->dx());
	GradientVec.y = (m_pVoxelLayer->color( i    , j + 1, k     ) - m_pVoxelLayer->color( i    , j - 1, k     )) / 255.0 / (2.0 * m_pVoxelLayer->dy());
	GradientVec.z = (m_pVoxelLayer->color( i    , j    , k + 1 ) - m_pVoxelLayer->color( i    , j    , k - 1 )) / 255.0 / (2.0 * m_pVoxelLayer->dz());
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CVoxelIntegrator::GetNormalizedGradientFromCache( int RegionNr, Vector& GradientVec )

// [6/12/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	for( auto Iter : m_GradientCache ) {
		if( Iter.first == RegionNr ) {
			GradientVec = Iter.second;
			return true;
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::AddGradientToCache( int RegionNr, const Vector& Gradient )

// [6/12/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	//m_pVisLayer->kernel.lock();
	m_pGeometry->lockWrite();


	// check if we have exceeded our cache size & remove the oldest one
	if( m_GradientCache.size() > cMaxGradientCacheSize ) {
		m_GradientCache.pop();
	}

	m_GradientCache.push( make_pair(RegionNr, Gradient) );

	//m_pVisLayer->kernel.unlock();
	m_pGeometry->unlockWrite();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//

bool CVoxelIntegrator::IdentifyVoxelRegion( Ray *pRay )

// Main driver of the ray marching routine
//
// Identify the voxel region of the ray with the passed position.
// In case we already have a region identified, we don't attempt the lookup, otherwise we try to identify it.
//
// [6/14/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// we already found a region (e.g. because the ray was passed at the first call of the shade() function in d3layer
	if( pRay->voxelreg >= 0 )
		return true;

	// Find yet undefined location
	pRay->voxelreg = m_pVoxelLayer->get( pRay->hit() );

	if( pRay->voxelreg >= 0 ) {
		pRay->normal = m_pVoxelLayer->voxel().normal( pRay->pos(), -pRay->dir() );
		return true;
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CVoxelIntegrator::GetShadingNormal( Ray* pRay )

// Determine the shading normal based on the gradient of the data
//
// FIXME: Verify if for the specular lighting we really need the ray's original normal. If not then we could replace the one
// carried by the ray and avoid recalculation + copies
//
// [6/14/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// for high quality renderings we interpolate also the normals
	if( m_RenderQuality > 2 )
		return CubicNormalInterpolation( pRay );

	return CalcNormalizedVoxelGradient( pRay );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::GetVoxelColor( Ray* pRay, Color& ShadingColor )

// determine the color of the specific voxel taking the rendermode (speed vs quality) into account
// [6/15/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if( m_RenderMode == CBaseIntegrator::cSpeed ) {
		ShadingColor.set( m_pVoxelLayer->color( pRay->voxelreg) );
		return;
	}

	// if we are in quality mode we resample the voxel data signal via interpolation
	if( m_RenderQuality == 2 )
		TrilinearVoxelInterpolation( pRay, ShadingColor );
	else
		CubicVoxelInterpolation( pRay, ShadingColor );			// Catmull - Rom, better results than trilinear and still sufficiently fast
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::TrilinearVoxelInterpolation( Ray* pRay, Color& ShadingColor )

// 3D trilinear voxel interpolation
// [6/15/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color	 CurColor;
	int		 VoxelIndex[3];	// i, j, k
	Vector VoxelCenter;
	Point  HitPt( pRay->hit() );

	// transform into voxel space where all other calculations will be done
	if( m_pVoxelLayer->voxel().hasMatrix() )
		m_pVoxelLayer->voxel().matrix().transform( HitPt );

	// get
	VoxelIndex[0] = m_pVoxelLayer->voxel().voxeli( HitPt.x );
	VoxelIndex[1] = m_pVoxelLayer->voxel().voxelj( HitPt.y );
	VoxelIndex[2] = m_pVoxelLayer->voxel().voxelk( HitPt.z );

	// transform the hit point into the local voxel's coordinates with (0,0,0) being the center of the voxel
	HitPt -= VoxelCenter;

	float Weight[3];

	// interpolate i
	for( int i = 0; i < 2; ++i ) {

		VoxelCenter.x = m_pVoxelLayer->voxel().voxelcx( VoxelIndex[0] + i );
		Weight[0] = std::fabs( 1.0 - std::fabs( HitPt.x - VoxelCenter.x ) / m_pVoxelLayer->dx());

		// interpolate j
		for( int j = 0; j < 2; ++j ) {

			VoxelCenter.y = m_pVoxelLayer->voxel().voxelcy( VoxelIndex[1] + j );
			Weight[1] = std::fabs( 1.0 - std::fabs( HitPt.y - VoxelCenter.y ) / m_pVoxelLayer->dy());

			// interpolate k
			for( int k = 0; k < 2; ++k ) {

				VoxelCenter.z = m_pVoxelLayer->voxel().voxelcz( VoxelIndex[2] + k );
				Weight[2] = std::fabs( 1.0 - std::fabs( HitPt.z - VoxelCenter.z ) / m_pVoxelLayer->dz());

				// carry out interpolation
				CurColor = m_pVoxelLayer->color( VoxelIndex[0] + i, VoxelIndex[1] + j, VoxelIndex[2] + k );
				ShadingColor += CurColor * Weight[0] * Weight[1] * Weight[2];

			} // end k
		} // end j
	} // end i

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::CubicVoxelInterpolation( Ray* pRay, Color& ShadingColor )

// 3D cubic spline voxel interpolation
//
// [6/16/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	const double B = 0, C = 0.5;			// Catmull - Rom spline
	Color	 CurColor;
	int		 VoxelIndex[3];	// i, j, k
	Vector VoxelCenter;
	Point	 HitPt( pRay->hit() );
	double TotalWeight = 0.0;


	// transform into voxel space where all other calculations will be done
	if( m_pVoxelLayer->voxel().hasMatrix() )
		m_pVoxelLayer->voxel().matrix().transform( HitPt );


	// get
	VoxelIndex[0] = m_pVoxelLayer->voxel().voxeli( HitPt.x );
	VoxelIndex[1] = m_pVoxelLayer->voxel().voxelj( HitPt.y );
	VoxelIndex[2] = m_pVoxelLayer->voxel().voxelk( HitPt.z );

	// transform the hit point into the local voxel's coordinates with (0,0,0) being the center of the voxel
	HitPt -= VoxelCenter;

	float Weight[3];

	// interpolate i
	for( int i = -1; i < 2; ++i ) {

		VoxelCenter.x = m_pVoxelLayer->voxel().voxelcx( VoxelIndex[0] + i );
		double X = std::fabs( HitPt.x - VoxelCenter.x ) / m_pVoxelLayer->dx();
		if( X < 0.5 )
			Weight[0] = ((12 - 9*B - 6*C) * X * X * X + (-18 + 12*B + 6*C) * X * X + (6 - 2*B)) / 6.0;
		else
			Weight[0] = ((-B - 6*C) * X * X * X + (6*B + 30*C) * X * X + (-12*B - 48*C) * X + (8*B + 24*C)) / 6.0;

		// interpolate j
		for( int j = -1; j < 2; ++j ) {

			VoxelCenter.y = m_pVoxelLayer->voxel().voxelcy( VoxelIndex[1] + j );
			double Y = std::fabs( HitPt.y - VoxelCenter.y ) / m_pVoxelLayer->dy();

			if( Y < 0.5  )
				Weight[1] = ((12 - 9 * B - 6 * C) * Y * Y * Y + (-18 + 12 * B + 6 * C) * Y * Y + (6 - 2 * B)) / 6.0;
			else
				Weight[1] = ((-B - 6 * C) * Y * Y * Y + (6 * B + 30 * C) * Y * Y + (-12 * B - 48 * C) * Y + (8 * B + 24 * C)) / 6.0;

			// interpolate k
			for( int k = -1; k < 2; ++k ) {

				VoxelCenter.z = m_pVoxelLayer->voxel().voxelcz( VoxelIndex[2] + k );
				double Z = std::fabs( HitPt.z - VoxelCenter.z ) / m_pVoxelLayer->dz();

				if( Z < 0.5 )
					Weight[2] = ((12 - 9 * B - 6 * C) * Z * Z * Z + (-18 + 12 * B + 6 * C) * Z * Z + (6 - 2 * B)) / 6.0;
				else
					Weight[2] = ((-B - 6 * C) * Z * Z * Z + (6 * B + 30 * C) * Z * Z + (-12 * B - 48 * C) * Z + (8 * B + 24 * C)) / 6.0;

				// carry out interpolation
				CurColor = m_pVoxelLayer->color( VoxelIndex[0] + i, VoxelIndex[1] + j, VoxelIndex[2] + k );
				ShadingColor += CurColor * Weight[0] * Weight[1] * Weight[2];
				TotalWeight += Weight[0] * Weight[1] * Weight[2];
			} // end k
		} // end j
	} // end i

	ShadingColor /= TotalWeight;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CVoxelIntegrator::CubicNormalInterpolation( Ray* pRay )

// 3D cubic spline interpolation of normals based on the voxel gradients
//
// [6/16/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	const double B = 0, C = 0.5;			// Catmull - Rom spline
	Vector CurNormal, InterpolNormal;
	double TotalWeight = 0;
	int		 VoxelIndex[3];	// i, j, k
	Vector VoxelCenter;
	Point  HitPt( pRay->hit() );

	// transform into voxel space where all other calculations will be done
	if( m_pVoxelLayer->voxel().hasMatrix() )
		m_pVoxelLayer->voxel().matrix().transform( HitPt );

	// get
	VoxelIndex[0] = m_pVoxelLayer->voxel().voxeli( HitPt.x );
	VoxelIndex[1] = m_pVoxelLayer->voxel().voxelj( HitPt.y );
	VoxelIndex[2] = m_pVoxelLayer->voxel().voxelk( HitPt.z );

	// transform the hit point into the local voxel's coordinates with (0,0,0) being the center of the voxel
	HitPt -= VoxelCenter;

	float Weight[3];

	// interpolate i
	for( int i = -1; i < 2; ++i ) {

		VoxelCenter.x = m_pVoxelLayer->voxel().voxelcx( VoxelIndex[0] + i );
		double X = std::fabs( HitPt.x - VoxelCenter.x ) / m_pVoxelLayer->dx();
		if( X < 0.5 )
			Weight[0] = ((12 - 9 * B - 6 * C) * X * X * X + (-18 + 12 * B + 6 * C) * X * X + (6 - 2 * B)) / 6.0;
		else
			Weight[0] = ((-B - 6 * C) * X * X * X + (6 * B + 30 * C) * X * X + (-12 * B - 48 * C) * X + (8 * B + 24 * C)) / 6.0;

		// interpolate j
		for( int j = -1; j < 2; ++j ) {

			VoxelCenter.y = m_pVoxelLayer->voxel().voxelcy( VoxelIndex[1] + j );
			double Y = std::fabs( HitPt.y - VoxelCenter.y ) / m_pVoxelLayer->dy();

			if( Y < 0.5 )
				Weight[1] = ((12 - 9 * B - 6 * C) * Y * Y * Y + (-18 + 12 * B + 6 * C) * Y * Y + (6 - 2 * B)) / 6.0;
			else
				Weight[1] = ((-B - 6 * C) * Y * Y * Y + (6 * B + 30 * C) * Y * Y + (-12 * B - 48 * C) * Y + (8 * B + 24 * C)) / 6.0;

			// interpolate k
			for( int k = -1; k < 2; ++k ) {

				VoxelCenter.z = m_pVoxelLayer->voxel().voxelcz( VoxelIndex[2] + k );
				double Z = std::fabs( HitPt.z - VoxelCenter.z ) / m_pVoxelLayer->dz();

				if( Z < 0.5 )
					Weight[2] = ((12 - 9 * B - 6 * C) * Z * Z * Z + (-18 + 12 * B + 6 * C) * Z * Z + (6 - 2 * B)) / 6.0;
				else
					Weight[2] = ((-B - 6 * C) * Z * Z * Z + (6 * B + 30 * C) * Z * Z + (-12 * B - 48 * C) * Z + (8 * B + 24 * C)) / 6.0;

				// carry out interpolation
				CalcVoxelGradient( VoxelIndex[0] + i, VoxelIndex[1] + j, VoxelIndex[2] + k, CurNormal );
				CurNormal.normalize();
				InterpolNormal += CurNormal * Weight[0] * Weight[1] * Weight[2];
				TotalWeight += Weight[0] * Weight[1] * Weight[2];
			} // end k
		} // end j
	} // end i

	InterpolNormal /= TotalWeight;
	InterpolNormal.normalize();
	return InterpolNormal;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CVoxelIntegrator::ExportColorHisto( const std::string& FileName )

// Export a histogram how many voxels use which color. This is useful if one wants
// to manually constract an opacity transfer function without GUI.
//
// [6/16/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	map<dword, int> ColorHisto;
	ofstream				Out( FileName );

	if( !Out )
		return false;

	assert( m_pVoxelLayer );
	m_pVoxelLayer->colorHistogram( ColorHisto );

	for( const auto& Iter : ColorHisto ) {
		Color32 CurCol = Iter.first;
		Out << CurCol.iR() << "\t" << CurCol.iG() << "\t" << CurCol.iB() << "\t" << Iter.second << endl;
	}

	Out.close();
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::SetRenderMode( int RenderMode, int  QualityLevel )

// define render mode & adjust ambient normalization accordingly. Higher quality render mode activates also interpolation
// of the normals (derivatives of the data) and thus, we need a bit more light --> smaller ambient normalization
//
// [6/24/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	CBaseIntegrator::SetRenderMode( RenderMode, QualityLevel );

	QualityLevel > 2 ? m_AmbientNormalization = 40 : m_AmbientNormalization = 75;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CVoxelIntegrator::CalcAmbientOcclusion( GeometryEngine* pEngine, Ray* pRay )

// calculate ambient occlusion for the current voxel
// [6/24/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double RandNr1, RandNr2, AOCDist = 0.0;
	Vector SampleDir;
	Ray		 AOCRay( *pRay );
	bool	 bScramble		= true;
	int    Scramble1		= bScramble ? pEngine->random.integer() : 0;
	int    Scramble2		= bScramble ? pEngine->random.integer() : 0;
	int		 NrOfSamples	= m_NrOfAOCSamples, ShadowedSamples = 0;

	// TODO: parallelize this
	for( int i = 0; i < NrOfSamples; i++ ) {
		RandNr1 = CBaseSampler::VanDerCorput( i, Scramble1 );
		RandNr2 = CBaseSampler::Sobol2( i, Scramble2 );

		SampleDir = CBaseSampler::uniformSampleHemisphere( GetShadingNormal( pRay ), RandNr1, RandNr2 );

		AOCRay = *pRay;
		AOCRay.segment().dir = SampleDir;
		AOCRay.segment().pos = pRay->hit( -2 * m_AOCRayMarchDelta );
		AOCRay.voxelreg = -1;
		AOCRay.depth = 0;
		AOCRay.segment().tmin = SMALL3D;	// ensure that we are out of the current (opaque) voxel

		AOCDist = MarchToNextOpaqueVoxel( &AOCRay );
		if( AOCDist < m_AOCMaxDist ) {
			ShadowedSamples++;
		}
	}

	// we artificially boost the LAOC by a factor of 2 to increase the 3D effect
	//return 1.0 - std::min( 1.0, 2.0 * ShadowedSamples / (NrOfSamples * 1.0));

	return m_AOCMinIntensity + (m_AOCMaxIntensity-  m_AOCMaxIntensity) * (NrOfSamples - ShadowedSamples) / (float)NrOfSamples;

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CVoxelIntegrator::MarchToNextOpaqueVoxel( Ray* pRay )

// march along the passed ray to the next non-transparent object and return the distance
// [6/24/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// quit if we're outside
	if( !IdentifyVoxelRegion( pRay ) )
		return 0.0;

	// or if we found an opaque object
	Color VoxelTransparency = RayToTransparency( pRay );
	if( VoxelTransparency < m_TransmissionThreshold )
		return 0.0;

	pRay->segment().tmin += m_AOCRayMarchDelta;	// march forward
	if( pRay->segment().tmin > m_AOCMaxDist )
		return 0.0;

	pRay->voxelreg = -1;
	pRay->incDepth();

	// continue marching
	MarchToNextOpaqueVoxel( pRay );

	return pRay->segment().tmin;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CVoxelIntegrator::IntegrateTransparency( Ray* pRay, Color& IntegratedTransparency )

// march along the passed ray to the next non-transparent object, integrate the transparency along the ray and return the distance
//
// NOTE: The primary transparency must be Color(1.0, 1.0, 1.0)!!
//
// [11/02/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// quit if we're outside
	if( !IdentifyVoxelRegion( pRay ) )
		return;

	// or if we found an opaque object
	IntegratedTransparency *= RayToTransparency( pRay );
	if( IntegratedTransparency < m_TransmissionThreshold )
		return;

	pRay->segment().tmin += m_AOCRayMarchDelta;	// march forward
	if( pRay->segment().tmin > pRay->segment().tmax )
		return;

	pRay->voxelreg = -1;
	pRay->incDepth();

	// continue marching
	IntegrateTransparency( pRay, IntegratedTransparency );

}

// vim:ts=2
