/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "arbbody.h"
#include "mesh.h"

// ================================= GARBBody ==================================
// FIXME position, get, Mesh...
void GARBBody::set(double* what)
{
	char err[128];
	err[0] = 0;

	GBody::createMesh();	// always create a mesh
	resetBBox();
	round0(what, nwhats(), SMALL);
	_mesh->allocateVertices(8);

	// add vertices
	P.set(what[0], what[1], what[2]);
	for (int i=0; i<8; i++)
		_mesh->vertex(i) = Point(what[i*3], what[i*3+1], what[i*3+2]);

	// copy faces
	bool same = true;
	for (int j=0,i=3*8; i<3*8+6; i++,j++) {
		int face = (int)what[i];
		if (face != faces[j]) {
			same = false;
			faces[j] = face;
		}

		int v = face%10 - 1;
		int zero = 0;
		if (v>=8) {
			sprintf(err, "Face %d contains an invalid vertex %d",(int)what[i],v);
			v = 0;
		}
		if (v==0) zero++;
		face /= 10;

		v = face%10 - 1;
		if (v>=8) {
			sprintf(err,"Face %d contains an invalid vertex %d",(int)what[i],v);
			v = 0;
		}
		if (v==0) zero++;
		face /= 10;

		v = face%10 - 1;
		if (v>=8) {
			sprintf(err,"Face %d contains an invalid vertex %d",(int)what[i],v);
			v = 0;
		}
		if (v==0) zero++;
		face /= 10;

		v = face%10 - 1;
		if (v>=8) {
			sprintf(err,"Face %d contains an invalid vertex %d",(int)what[i],v);
			v = 0;
		}
		if (v==0) zero++;

		if (zero>1) {
			sprintf(err,"Face %d contains too few vertices",(int)what[i]);
			continue;
		}
	}
	if (!same) {
		_mesh->freeFaces();
		_mesh->freeEdges();
	}
	if (err[0]) throw err;
} // set

/* --- get --- */
int GARBBody::get(double *what) const
{
	int w = 0;
	for (int i=0; i<8; i++) {
		what[w++] = _mesh->vertex(i).x;
		what[w++] = _mesh->vertex(i).y;
		what[w++] = _mesh->vertex(i).z;
	}
	for (int i=0; i<6; i++)
		what[w++] = (double)faces[i];
	return 30;
} // get

/** @return faceVertices */
bool GARBBody::faceVertices(int f, int v[4], Point* V[4])
{
	int face = faces[f];
	if (!face) {
		// The sizeof(v/V[0])*4 is needed since
		// sizeof(v/V) will return the size of int*/Point** not the size of the array
		memset(v,0,sizeof(v[0])*4);
		memset(V,0,sizeof(V[0])*4);
		return false;
	}
	for (int i=0; i<4; i++) {
		v[i] = face%10 - 1;
		if (v[i]>=8) v[i] = 0;
		if (v[i]>=0)
			V[i] = &_mesh->vertex(v[i]);
		else
			V[i] = NULL;
		face /= 10;
	}
	return true;
} // faceVertices

/** createQuads */
void GARBBody::createQuads()
{
	GBody::createQuads();
	// create faces and quadratics
	for (int i=0; i<6; i++) {
		int v[4];
		Point* V[4];
		faceVertices(i, v, V);
		if (!V[0] || !V[1] || !V[2]) continue;

		// create quads
		Vector N = (*V[1] - *V[0]) ^ (*V[2] - *V[0]);
		N.normalize();

		addQuad(-N.x, -N.y, -N.z, N*(*V[0]));

		if (V[3]!=NULL) {
			// check for co-planarity!
			if (!Eq0(_Q[_nQ-1](V[3]->x, V[3]->y, V[3]->z), SMALL))
				fprintf(stderr,"Face %d not planar\n",(int)faces[i]);
		}

		// Check remaining vertices, from which side they are
		bool first = true;
		bool positive = false;
		for (int j=0; j<8; j++) {
			if (j==v[0] || j==v[1] || j==v[2] || j==v[3]) continue;
			double q = _Q[_nQ-1](	_mesh->vertex(j).x,
						_mesh->vertex(j).y,
						_mesh->vertex(j).z);
			if (Eq0(q, SMALL))
				fprintf(stderr,"Vertex %d is coplanar with face %d\n",j+1,(int)faces[i]);
			if (first) {
				positive = (bool)(q>0.0);
				first = false;
			} else {
				if (positive != (bool)(q>0.0))
					fprintf(stderr,"Cannot resolve ARB orientation\n");
			}
		}
		// if remaining vertices are from same side with the normal
		if (positive)
			_Q[_nQ-1].negate();		// swap the sign of the quad
	}
} // createQuads

/* createMesh */
void GARBBody::createMesh()
{
	GBody::createMesh();
	if (_mesh->nedges()) return;

	for (int i=0; i<6; i++) {
		int v[4];
		Point* V[4];
		if (!faceVertices(i, v, V)) continue;
		if (!V[0] || !V[1] || !V[2]) continue;
		bool last = (bool)(v[3]>=0);
		_mesh->add(v[0],v[1],v[2], true, true, !last);
		if (last)
			_mesh->add(v[2],v[3],v[0], true, true, false);
	}
	_mesh->calcBbox();
	_mesh->process();
	if (!_mesh->isClosed()) {
		fprintf(stderr,"ERROR open ARB a face is missing\n");
		return;
	}
	if (!_mesh->isOrientable())
		_mesh->makeOrientable();

	if (_mesh->volume()<0.) _mesh->flip();

	assert(_mesh->isClosed());
	assert(_mesh->isOrientable());
#if _DEBUG>2
	cout << endl;
	cout << "ARB Mesh " << endl;
	cout << mesh << endl;
	cout << " isClosed=" << _mesh->isClosed();
	cout << " isOrientable=" << _mesh->isOrientable() << endl;
	cout << "ARB volume=" << _mesh->volume() << endl;
#endif
} // createMesh

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GARBBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point dR = r - P;
	for (size_t i=0; i<_mesh->nvertices(); i++)
		_mesh->vertex(i) += dR;

	GBody::move(item, r, w);
} // move

/** rotate body around an axis and current body position
 * @param angle	to rotate
 * @param axis	to rotate
 */
void GARBBody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle, axis);
	Matrix4	rot;			// duplicate rot matrix with the GBody::rotate
	rot.rotate(angle, axis);
	for (int i=1; i<8; i++) {
		Vector v = _mesh->vertex(i) - P;
		_mesh->vertex(i) = rot*v + P;
	}
} // rotate

/** @return bounding box of body */
void GARBBody::updateBBox() const
{
	assert(_mesh!=nullptr);
	if (_mesh->nvertices()) {
		// use the vertices as a guess
		_bbox.add(position());
		_bbox.Union(_mesh->bbox());
	} else
		_bbox.infinite();
} // updateBBox

#ifdef OBBOX
/** @return oriented bounding box of body */
void GARBBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];
	if (!in) {
		if (_mesh->nvertices()) {
			// use the vertices as a guess
			bb.add(position());
			for (size_t i=0; i<_mesh->nvertices(); i++)
				bb.add(_mesh->vertex(i));
		} else
			bb.infinite();
	} else {
		// TODO: Implement inside obb?
		assert(false);
	}
} // updateOBB
#endif
