/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "cylbody.h"

#include "mesh.h"
#include "line.h"

// =============================== GCylinderBase ===============================
int GCylinderBase::N_CYLINDER = 16;

/** make cone quadrics from 3 vectors and 2 positions */
void GCylinderBase::makeConeQuads()
{
	Matrix4 M, Minv, T;

	Minv.make(X,Y,Z);
	Minv.inverse();
	T.translate(P);
	M.multiply(T, Minv);
	Minv.inverse(M);

	// transform first quad
	_Q[0].transform(Minv, M);
//	_Q[0].normalize();

	// two planes for the caps
	//        Cx     Cy     Cz      C
	addQuad(-Z.x,-Z.y,-Z.z, Z*P);	// base
	addQuad( Z.x, Z.y, Z.z,-Z*Po);	// apex
} // makeConeQuads

/** createEllConeMesh - create mesh of an elliptical cone
 * @param Rbx	base radius along X vector
 * @param Rby	base radius along Y vector
 * @param Rax	apex radius along X vector
 * @param Ray	apex radius along Y vector
 */
void GCylinderBase::createEllConeMesh(	const double Rbx, const double Rby,
					const double Rax, const double Ray,
					const bool isinfinite)
{
	GBody::createMesh();
	bool first = _mesh->isEmpty();
	const double step = PI2 / (double)N_CYLINDER;
	double ang = 0.0;

	Point A, B, C;
	if (isinfinite) {
		B = P - infinite*Z;
		A = P + infinite*Z;
		C = P;		// additional ring!
	} else {
		B = P;
		A = Po;
	}

	if (first) _mesh->allocateVertices(2+2*N_CYLINDER);

	_mesh->vertex(0) = B;	// base
	_mesh->vertex(1) = A;	// apex

	// Build base
	//if (!Eq0(Rbx,SMALL) and !Eq0(Rby,SMALL))	// ignore Cones with tip
	//if (!Eq0(Rax,SMALL) and !Eq0(Ray,SMALL))	// ignore Cones with tip

	const int base = 2;
	const int apex = 2 + N_CYLINDER;

	for (int i=0; i<N_CYLINDER; i++, ang+=step) {
		double c, s;
		bsincos(ang, &s, &c);
		_mesh->vertex(i+base) = B + (Rbx*c)*X + (Rby*s)*Y;
		_mesh->vertex(i+apex) = A + (Rax*c)*X + (Ray*s)*Y;
	}
	_mesh->calcBbox();

	if (first) {
		// XXX I could do in a single loop
		// but this way I could potentially use the fan opengl commands

		// create base faces
		for (int i=0; i<N_CYLINDER; i++) {
			int in = (i==N_CYLINDER-1? 0 : i+1); // next i
			_mesh->add(0, base+in, base+i, false, true, false);
		//}

		// create apex faces
		//for (int i=0; i<N_CYLINDER; i++) {
		//	int in = (i==N_CYLINDER-1? 0 : i+1); // next i
			_mesh->add(1, apex+i, apex+in, false, true, false);
		//}

		// create side faces
		//for (int i=0; i<N_CYLINDER; i++) {
		//	int in = (i==N_CYLINDER-1? 0 : i+1); // next i
			_mesh->add(base+i, apex+in, apex+i,  false, true, true);
			_mesh->add(base+i, base+in, apex+in, true,  true, false);
		}

		_mesh->process();
		assert(_mesh->isClosed());
		assert(_mesh->isOrientable());
#if _DEBUG>2
		cout << endl;
		cout << "CYL Mesh ";
		cout << " isClosed=" << _mesh->isClosed();
		cout << " isOrientable=" << _mesh->isOrientable() << endl;
		cout << "CYL volume=" << _mesh->volume() << endl;
#endif
	}
} // createEllConeMesh

/** bbox_addRotatedEllipse
 * add to the bounding box an arbitrary rotated ellipse
 */
static void bbox_addRotatedEllipse(BBox& bb,
		const Vector& rotX, const Vector& rotY,
		const Point& pos, double ra, double rb)
{
	for (int i = 0; i < 3; i++) {
		double a = ra * rotX[i];
		double b = rb * rotY[i];
		double a2b2 = a*a + b*b;

		if (a2b2 < SMALL) continue;

		double sint = sqrt(b*b/a2b2);
		double cost = 0.0;
		if (Abs(sint) > SMALL)
			cost = a / b * sint;
		else {
			cost = 1.0;
			sint = 0.0;
		}

		Point e;
		e.x = rotX.x * ra * cost + rotY.x * rb * sint;
		e.y = rotX.y * ra * cost + rotY.y * rb * sint;
		e.z = rotX.z * ra * cost + rotY.z * rb * sint;

		bb.add(pos.x+e.x, pos.y+e.y, pos.z+e.z);
		bb.add(pos.x-e.x, pos.y-e.y, pos.z-e.z);
	}
} // bbox_addRotatedEllipse

// ================================= GRCCBody ==================================
/** set */
void GRCCBody::set(const double x,  const double y,  const double z,
		   const double hx, const double hy, const double hz,
		   const double r)
{
	P.set( x,  y,  z);
	Z.set(hx, hy, hz);
	R(Max(r,0.0));

	Po = P + Z;
	zlen = Z.normalize();

	findXYZ();
	if (r<TOOSMALL) throw std::string("Negative or Zero radius");
} // set

/** set */
void GRCCBody::set(double* what)
{
	resetBBox();
	round0(what, nwhats(), SMALL2);
	set(what[0], what[1], what[2],
	    what[3], what[4], what[5],
	    what[6]);
} // set

/* --- get --- */
int GRCCBody::get(double* what) const
{
	what[0] = P.x;
	what[1] = P.y;
	what[2] = P.z;
	what[3] = zlen*Z.x;
	what[4] = zlen*Z.y;
	what[5] = zlen*Z.z;
	what[6] = R();
	return 7;
} // get

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GRCCBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= SP;

	switch (item) {
		case 1:	// Find distance from Z-axis
			R(Sqrt(tr.length2() - Sq(tr*Z)));
			break;

		case 2: {	// Move base
				double zp = tr*SZ;
				P = SP + zp*SZ;
				zlen = szlen - zp;
				if (zlen<0.0) {
					Z = -SZ;
					zlen = -zlen;
				} else
					Z = SZ;
			}
			break;

		case 3:		// Move apex
			zlen = tr*SZ;
			if (zlen<0.0) {
				Z = -SZ;
				zlen = -zlen;
			} else
				Z = SZ;
			Po = SP + zlen*Z;
			break;

		default:
			GBody::move(item, r, w);
	}
} // move

/** createQuads */
void GRCCBody::createQuads()
{
	GBody::createQuads();
	addQuad(Quad(Quad::Type::CYLINDER, P, Z, R()));
	addQuad(-Z.x,-Z.y,-Z.z, Z*P);	// base
	addQuad( Z.x, Z.y, Z.z,-Z*Po);	// apex
} // createQuads

/** @return bounding box of body */
void GRCCBody::updateBBox() const
{
	Point p = position();
	Point po = p + (vectorZ()*zlen);

	_bbox.add(p);
	_bbox.add(po);

	bbox_addRotatedEllipse(_bbox, vectorX(), vectorY(), p,  R(), R());
	bbox_addRotatedEllipse(_bbox, vectorX(), vectorY(), po, R(), R());
} // bbox

#ifdef OBBOX
/** update oriented bounding box of body */
void GRCCBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	if (!in) {
		bb.low()  = Point(-R(), -R(), 0.0);
		bb.high() = Point( R(),  R(), zlen);
	} else {
		double sc = R()/Sqrt2;
		bb.low()  = Point(-sc, -sc, 0.0);
		bb.high() = Point( sc,  sc, zlen);
	}
} // updateOBB
#endif

// ================================= GRECBody ==================================
/** set */
void GRECBody::set(const double x,  const double y,  const double z,
		   const double wx, const double wy, const double wz,
		   const double ux, const double uy, const double uz,
		   const double vx, const double vy, const double vz)
{
	P.set( x,  y,  z);
	Z.set(wx, wy, wz);
	X.set(ux, uy, uz);
	Y.set(vx, vy, vz);

	Po = P + Z;

	// correct for perpendicularity
	xlen = X.normalize();
	ylen = Y.normalize();
	zlen = Z.normalize();

	if (xlen<TOOSMALL) throw std::string("X-length is zero");
	if (ylen<TOOSMALL) throw std::string("Y-length is zero");
	checkOrthogonal();
} // set

/** set */
void GRECBody::set(double* what)
{
	resetBBox();
	round0(what, nwhats(), SMALL2);
	set(what[0], what[ 1], what[ 2],
	    what[3], what[ 4], what[ 5],
	    what[6], what[ 7], what[ 8],
	    what[9], what[10], what[11]);
} // set

/* --- get --- */
int GRECBody::get(double* what) const
{
	what[ 0] = P.x;
	what[ 1] = P.y;
	what[ 2] = P.z;
	what[ 3] = zlen*Z.x;
	what[ 4] = zlen*Z.y;
	what[ 5] = zlen*Z.z;
	what[ 6] = xlen*X.x;
	what[ 7] = xlen*X.y;
	what[ 8] = xlen*X.z;
	what[ 9] = ylen*Y.x;
	what[10] = ylen*Y.y;
	what[11] = ylen*Y.z;
	return 12;
} // get

/**  @return closest point, edge, face */
int GRECBody::closest(const Point& r, const double d, const Vector& w) const
{
	int close = GBody::closest(r,d,w);
	if (close == 1) {
		double px = Abs(Y*(r-P));
		double py = Abs(X*(r-P));
		if (px>py) return 4;
	}
	return close;
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GRECBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= SP;
	switch (item) {
		case 1:	{ // X-radius
				// Find a solution where the surface cross the r
				Point rp = P + (tr*Z)*Z;	// projection on Z axis
				double x = (r-rp)*X;		// x,y components on X,Y axis
				double y = (r-rp)*Y;
				// equation
				//  x^2    y^2
				//  ---  + ---  = 1
				//  xl^2   yl^2
				double d = 1.0 - Sq(y)/Sq(sylen);
				if (d > SMALL)
					xlen = sqrt(Sq(x) / d);
				else	// On failure find distance from Y-axis
					xlen = Sqrt(tr.length2() - Sq(tr*Y));
			}
			break;

		case 2: {	// Move base
				double zp = tr*SZ;
				P = SP + zp*SZ;
				zlen = szlen - zp;
				if (zlen<0.0) {
					Z = -SZ;
					zlen = -zlen;
				} else
					Z = SZ;
			}
			break;

		case 3:		// Move apex
			zlen = tr*SZ;
			if (zlen<0.0) {
				Z = -SZ;
				zlen = -zlen;
			} else
				Z = SZ;
			Po = P + zlen*Z;
			break;

		case 4:	{ // X-radius
				// Find distance from Z-axis
				Point rp = P + (tr*Z)*Z;	// projection on Z axis
				double x = (r-rp)*X;	// x,y components on X,Y axis
				double y = (r-rp)*Y;
				double d = 1.0 - Sq(x)/Sq(sxlen);
				if (d > SMALL)
					ylen = Sqrt(Sq(y) / d);
				else	// On failure find distance from X-axis
					ylen = Sqrt(tr.length2() - Sq(tr*X));
			}
			break;

		default:
			GBody::move(item, r, w);
	}
} // move

/** createQuads */
void GRECBody::createQuads()
{
	GBody::createQuads();
	if (ISZERO(xlen) || ISZERO(ylen)) return;

	// Create the cylinder around Z then transform
	//  Cxx Cyy Czz  Cxy  Cxz  Cyz  Cx   Cy   Cz    C
	addQuad( 1.0/Sq(xlen),  1.0/Sq(ylen), 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,-1.0);
	makeConeQuads();
} // createQuads

/** @return bounding box of body */
void GRECBody::updateBBox() const
{
	Point p = position();
	Point po = p + (vectorZ()*zlen);

	_bbox.add(p);
	_bbox.add(po);

	bbox_addRotatedEllipse(_bbox, vectorX(), vectorY(), p,  xlen, ylen);
	bbox_addRotatedEllipse(_bbox, vectorX(), vectorY(), po, xlen, ylen);
} // bbox

#ifdef OBBOX
/** update oriented bounding box of body */
void GRECBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	if (!in) {
		bb.low()  = Point(-xlen, -ylen, 0.0);
		bb.high() = Point( xlen,  ylen, zlen);
	} else {
		double sc = 1.0/Sqrt2;
		bb.low()  = Point(-xlen*sc, -ylen*sc, 0.0);
		bb.high() = Point( xlen*sc,  ylen*sc, zlen);
	}
} // updateOBB
#endif

// ================================= GTRCBody ==================================
/** set */
void GTRCBody::set(const double x,  const double y,  const double z,
		   const double hx, const double hy, const double hz,
		   const double rb, const double ra)
{
	P.set( x,  y,  z);
	Z.set(hx, hy, hz);

	if (rb<0.0) throw std::string("Negative base radius");
	if (ra<0.0) throw std::string("Negative apex radius");
	if (ra<TOOSMALL && rb<TOOSMALL) throw std::string("Both apex and base radius are too small");
	Rb(Max(rb,0.0));	// Rb base radius
	Ra(Max(ra,0.0));	// Ra apex radius

	Po   = P + Z;
	h2   = Z.length2();	// to keep precision
	zlen = Z.normalize();
	if (zlen<TOOSMALL) throw std::string("Zero height");

	findXYZ();
} // set

/** set */
void GTRCBody::set(double* what)
{
	resetBBox();
	round0(what, nwhats(), SMALL2);
	set( what[0], what[1], what[2],
	     what[3], what[4], what[5],
	     what[6], what[7]);
} // set

/* --- get --- */
int GTRCBody::get(double* what) const
{
	what[0] = P.x;
	what[1] = P.y;
	what[2] = P.z;
	what[3] = zlen*Z.x;
	what[4] = zlen*Z.y;
	what[5] = zlen*Z.z;
	what[6] = Rb();
	what[7] = Ra();
	return 8;
} // get

/**  @return closest point, edge, face */
int GTRCBody::closest(const Point& r, const double d, const Vector& w) const
{
	int close = GBody::closest(r,d,w);
	if (close == 1) {
		// Check which is closer, base or apex?
		if (Abs(Q(1)(r)) > Abs(Q(2)(r)))
			return 4;
	}
	return close;
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GTRCBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= SP;
	switch (item) {
		case 1:	// Find distance from Z-axis
			Rb(Sqrt(tr.length2() - Sq(tr*Z)));	// base radius
			break;

		case 2: {	// Move base
				double zp = tr*SZ;
				P = SP + zp*SZ;
				zlen = szlen - zp;
				if (zlen<0.0) {
					Z = -SZ;
					zlen = -zlen;
				} else
					Z = SZ;
				h2 = Sq(zlen);
			}
			break;

		case 3:		// Move apex
			zlen = tr*SZ;
			if (zlen<0.0) {
				Z = -SZ;
				zlen = -zlen;
			} else
				Z = SZ;
			h2 = Sq(zlen);
			Po = P + zlen*Z;
			break;

		case 4:		// Change apex radius
			Ra(Sqrt(tr.length2() - Sq(tr*Z)));
			break;

		default:
			GBody::move(item, r, w);
			h2 = Sq(zlen);
	}
} // move

/** createQuads */
void GTRCBody::createQuads()
{
	GBody::createQuads();

	if (ISZERO(zlen)) return;

	// Create the cone around Z then transform
	//  Cxx Cyy Czz  Cxy  Cxz  Cyz  Cx   Cy   Cz    C
//	double s = Ra()-Rb();	// slope=(Ra-Rb)/zlen but avoid division that generates roundup errors
//	addQuad(1.0, 1.0, -s*s/h2, 0.0, 0.0, 0.0, 0.0, 0.0,-2.0*s*Rb()/zlen, -Rb()*Rb());
//	makeConeQuads();
	addQuad(Quad(Quad::Type::CONE, P, zlen*Z, Rb(), Ra()));
	addQuad(-Z.x,-Z.y,-Z.z, Z*P);	// base
	addQuad( Z.x, Z.y, Z.z,-Z*Po);	// apex
} // createQuads

/** @return bounding box */
void GTRCBody::updateBBox() const
{
	Point p  = position();
	Point po = p + zlen*vectorZ();

	_bbox.add(p);
	_bbox.add(po);

	bbox_addRotatedEllipse(_bbox, vectorX(), vectorY(), p,  Rb(), Rb());
	bbox_addRotatedEllipse(_bbox, vectorX(), vectorY(), po, Ra(), Ra());
} // bbox

#ifdef OBBOX
/** update oriented bounding box of body */
void GTRCBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();

	if (!in) {
		double R = Max(Ra(), Rb());
		bb.low()  = Point(-R, -R, 0.0);
		bb.high() = Point( R,  R, zlen);
	} else {
		double sc, z;
		if (Rb() > Ra()) {
			sc = Rb() / 2.0;
			if (Ra() >= sc) {
				sc = Ra();
				z  = zlen;
			} else
				z  = sc*zlen/(Rb()-Ra());
			sc /= Sqrt2;
			bb.low()  = Point(-sc, -sc, 0.0);
			bb.high() = Point( sc,  sc, z);
		} else {
			sc = Ra() / 2.0;
			if (Rb() >= sc) {
				sc = Rb();
				z  = 0.0;
			} else
				z = zlen - sc*zlen/(Ra()-Rb());
			sc /= Sqrt2;
			bb.low()  = Point(-sc, -sc, z);
			bb.high() = Point( sc,  sc, zlen);
		}
	}
} // updateOBB
#endif

// ============================== GInfEllCylBody ===============================
/** set */
void GInfEllCylBody::set(const double a, const double b, const double ra, const double rb)
{
	if (ra<TOOSMALL or rb<-TOOSMALL)
		throw std::string("Negative radius");

	Rx(Max(ra,0.0));
	Ry(Max(rb,0.0));

	switch (type()) {
		case Type::XCC:
		case Type::XEC:
			P.set(0.0, a, b);
			if (type()==Type::XCC) ylen = xlen;
			Z = Vector::Xo;
			X = Vector::Yo;
			Y = Vector::Zo;
			break;
		case Type::YCC:
		case Type::YEC:
			P.set(b, 0.0, a);
			if (type()==Type::YCC) ylen = xlen;
			Z = Vector::Yo;
			X = Vector::Zo;
			Y = Vector::Xo;
			break;
		case Type::ZCC:
		case Type::ZEC:
			P.set(a, b, 0.0);
			if (type()==Type::ZCC) ylen = xlen;
			X = Vector::Xo;
			Y = Vector::Yo;
			Z = Vector::Zo;
			break;
		default:
			assert(false);
	}
} // set

/** set */
void GInfEllCylBody::set(double* what)
{
	resetBBox();
	round0(what, nwhats(), SMALL2);
	set(what[0], what[1], what[2], what[3]);
} // set

/* --- get --- */
int GInfEllCylBody::get(double* what) const
{
	what[2] = Rx();
	what[3] = Ry();		// For [XYZ]EC only

	switch (type()) {
		case Type::XCC:
		case Type::XEC:
			what[0] = P.y;
			what[1] = P.z;
			if (type()==Type::XEC) return 4;
			return 3;

		case Type::YCC:
		case Type::YEC:
			what[0] = P.z;
			what[1] = P.x;
			if (type()==Type::YEC) return 4;
			return 3;

		case Type::ZCC:
		case Type::ZEC:
			what[0] = P.x;
			what[1] = P.y;
			if (type()==Type::ZEC) return 4;
			return 3;

		default: return -1;
	}
} // get

/**  @return closest point, edge, face */
int GInfEllCylBody::closest(const Point& r, const double d, const Vector& w) const
{
	int close = GBody::closest(r,d,w);

	switch (type()) {
		case Type::XEC:
			if (Abs(r.y-P.y) < Abs(r.z-P.z))
				return 2;
			return close;
		case Type::YEC:
			if (Abs(r.z-P.z) < Abs(r.x-P.x))
				return 2;
			return close;
		case Type::ZEC:
			if (Abs(r.x-P.x) < Abs(r.y-P.y))
				return 2;
			return close;
		default:
			return close;
	}
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GInfEllCylBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= SP;

	double x, y;
	bool ell = true;

	switch (type()) {
		case Type::XCC:
			ell = false;
			/* FALLTHRU */
		case Type::XEC:
			x = tr.y;
			y = tr.z;
			break;

		case Type::YCC:
			ell = false;
			/* FALLTHRU */
		case Type::YEC:
			x = tr.z;
			y = tr.x;
			break;

		case Type::ZCC:
			ell = false;
			/* FALLTHRU */
		case Type::ZEC:
			x = tr.x;
			y = tr.y;
			break;

		default:
			x = y = 0.0;	// Never comes here but keeps compiler happy
			assert(false);
	}

	if (item==1) {
		if (ell) {
			double d = 1.0 - Sq(y)/Sq(ylen);
			if (d > SMALL)
				Rx(Sqrt(Sq(x) / d));
			else	// On failure find distance from Y-axis
				Rx(x);
		} else
			R(Sqrt(Sq(x) + Sq(y)));
	} else
	if (item==2) {
		double d = 1.0 - Sq(x)/Sq(Rx());
		if (d > SMALL)
			ylen = Sqrt(Sq(y) / d);
		else
			ylen = y;
	} else
		GBody::move(item, r, w);
} // move

/** rotate body around an axis */
void GInfEllCylBody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle, axis);

	/* find closest axis to align to */
	double ax = Abs(Z.x);
	double ay = Abs(Z.y);
	double az = Abs(Z.z);

	if (az>ax && az>ay) {
		Z = Vector::Zo;
		if (InRange(Type::XEC, type(), Type::ZEC))
			_type = Type::ZEC;
		else
			_type = Type::ZCC;
	} else
	if (ay>ax && ay>az) {
		Z = Vector::Yo;
		if (InRange(Type::XEC, type(), Type::ZEC))
			_type = Type::YEC;
		else
			_type = Type::YCC;
	} else {
		Z = Vector::Zo;
		if (InRange(Type::XEC, type(), Type::ZEC))
			_type = Type::XEC;
		else
			_type = Type::XCC;
	}

	findXYZ();
} // rotate

/** createQuads */
void GInfEllCylBody::createQuads()
{
	GBody::createQuads();
	double a = 1.0, b = 1.0, c = 1.0;
	double r2 = Sq(xlen);

	switch (type()) {
		case Type::XCC:
			addQuad(Quad(Quad::Type::CYLINDER_X, P, R()));
			return;
		case Type::YCC:
			addQuad(Quad(Quad::Type::CYLINDER_Y, P, R()));
			return;
		case Type::ZCC:
			addQuad(Quad(Quad::Type::CYLINDER_Z, P, R()));
			return;
		default:
//			assert(0);

	if (type()==Type::XCC || type()==Type::XEC) {
		a = 0.0;
		if (type()==Type::XEC) {
			c   = r2;
			b   = Sq(ylen);
			r2 *= b;
		}
	} else
	if (type()==Type::YCC || type()==Type::YEC) {
		b = 0.0;
		if (type()==Type::YEC) {
			a   = r2;
			c   = Sq(ylen);
			r2 *= c;
		}
	} else
	if (type()==Type::ZCC || type()==Type::ZEC) {
		c = 0.0;
		if (type()==Type::ZEC) {
			b   = r2;
			a   = Sq(ylen);
			r2 *= a;
		}
	}
	//      Cxx Cyy Czz Cxy  Cxz  Cyz
	addQuad( a,  b,  c, 0.0, 0.0, 0.0,
	//            Cx            Cy            Cz
		-2.0*a*P.x,  -2.0*b*P.y,  -2.0*c*P.z,
		 a*Sq(P.x) + b*Sq(P.y) + c*Sq(P.z) - r2);
}
} // createQuads

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 */
GBody::Location GInfEllCylBody::_locationWrt(const GBody *body) const
{
//	if (body->type() == Type::RCC) {
//	} else
	if (type() == Type::XCC || type() == Type::YCC || type() == Type::ZCC) {
		if (body->type() == Type::XCC || body->type() == Type::YCC ||
		    body->type() == Type::ZCC) {
			Vector thisZ = vectorZ();
			Vector bodyZ = body->vectorZ();

			// Minimum distance between the two axes
			double dist = lineLineDistance(position(), thisZ, body->position(), bodyZ);

			// If they are parallel.
			// FIXME could be that after a rotation thisZ = +/- bodyZ
			if (type() == body->type()) {
				if (dist + ((const GInfEllCylBody*)body)->R() <  R())
					return Location::BinA;
				else
				if (dist - ((const GInfEllCylBody*)body)->R() < -R())
					return Location::AinB;
				else
				if (dist - ((const GInfEllCylBody*)body)->R() >  R())
					return Location::OUTSIDE;
				else
					return Location::OVERLAP;
			} else {
				// any other skew orientation (after rotation)
				if (dist - ((const GInfEllCylBody*)body)->R() >  R())
					return Location::OUTSIDE;
				else
					return Location::OVERLAP;
			}
		}
	}

	return GBody::_locationWrt(body);
} // _locationWrt

/** @return bounding box of body */
void GInfEllCylBody::updateBBox() const
{
	switch(type()) {
		case Type::XCC:
			_bbox.low(-INFINITE, P.y-R(), P.z-R());
			_bbox.high(INFINITE, P.y+R(), P.z+R());
			break;
		case Type::YCC:
			_bbox.low( P.x-R(), -INFINITE, P.z-R());
			_bbox.high(P.x+R(),  INFINITE, P.z+R());
			break;
		case Type::ZCC:
			_bbox.low( P.x-R(), P.y-R(), -INFINITE);
			_bbox.high(P.x+R(), P.y+R(),  INFINITE);
			break;

		case Type::XEC:
			_bbox.low(-INFINITE, P.y-Rx(), P.z-Ry());
			_bbox.high(INFINITE, P.y+Rx(), P.z+Ry());
			break;
		case Type::YEC:
			_bbox.low( P.x-Ry(), -INFINITE, P.z-Rx());
			_bbox.high(P.x+Ry(),  INFINITE, P.z+Rx());
			break;
		case Type::ZEC:
			_bbox.low( P.x-Rx(), P.y-Ry(), -INFINITE);
			_bbox.high(P.x+Rx(), P.y+Ry(),  INFINITE);
			break;

		default:
			assert(0);
	}
	if (_hasMatrix) _bbox.transform(_matrix);
} // bbox

#ifdef OBBOX
/** update oriented bounding box of body */
void GInfEllCylBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];
	bb.infinite();

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();

	if (!in) {
		bb.low().x  = -Rx();
		bb.low().y  = -Ry();
		bb.high().x =  Rx();
		bb.high().y =  Ry();
	} else {
		double sc = 1.0/Sqrt2;

		bb.low().x  = -Rx()*sc;
		bb.low().y  = -Ry()*sc;
		bb.high().x =  Rx()*sc;
		bb.high().y =  Ry()*sc;
	}
} // updateOBB
#endif
