/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	15-Mar-2012
 */

#ifndef __GEOMETRY_GZONE_H
#define __GEOMETRY_GZONE_H

#include <assert.h>

#include "gbody.h"
#include "math/vector.h"

class Quad;
class GRegion;

// =================================== GZone ===================================
/** Zone class. Zone is a region in space defined:
 * 1. STD (normal form) with only with intersection or subtraction
 * 2. RPN form with an Reverse Polish Notion expression
 *
 */
class GZone {
//	using GBodyArray = std::vector<GBody*>;
	using GBodyArray = Array<GBody*>;

// Zone expression max stack nesting level
static	const	int ZONE_STACK_SIZE = 100;

public:
	enum class Format : int {
		STD       = 0,	// Product expects [Plus-Bodies]... <NullBody> [Minus-Bodies]...
		PRODUCT   = 0,	// alias to STD
		RPN       = 1,	// (Postfix) Reverse polish notation with body, +, -, |, @(Universe)
		POSTFIX   = 1,	// operator follows operands
		NORMAL    = 2,	// (Infix)   Normal mathematical expression with parenthesis
		INFIX     = 2,	// operator inside operands
		//PREFIX  = 3,	// (Prefix) operator precedes operands
		//RPNEXP  = 4,	// RPN Expanded expression no nesting/parenthesis
		//		//	product product | product |..
		//NORMALEXP = 5,	// Normal expression expanded with no parenthesis
		//		//	product | product...
	};

private:
	int	_id;		// id of zone
	Format	_format;	// zone expression format
	int	_generation;	// Zone generation
	bool	_bodyref;	// bodies has reference to this
				// only edit zones should not have ref to bodies
	GBodyArray expr;	// expression or rpn
mutable dword	_hash;		// hash cache
mutable	BBox::State _bboxState;	// status of bbox
mutable	BBox	_bbox;		// cached information on bounding box
#ifdef OBBOX
mutable	BBox::State _obboxState;// status of obbox
mutable OBBox	_obbox;		// cached information on oriented bounding box
#endif

public:
	GRegion* region;	// parent region	// FIXME make private

public:
	GZone(GRegion* reg=nullptr, const bool ref=true);
	~GZone()				{ clear(); }

	void	clear(bool fast=false);

	// Id
	void	id(const int i)			{ _id = i; }
	int	id()			const	{ return _id; }
	std::string	name() const;

	/* set region mode RPN or STanDard */
	void	format(Format f)		{ _format = f; }
	Format	format()	const		{ return _format; }
	bool	std()		const		{ return _format==Format::PRODUCT; }
	bool	isproduct()	const		{ return _format==Format::PRODUCT; }
	bool	rpn()		const		{ return _format==Format::RPN;     }
	bool	postfix()	const		{ return _format==Format::POSTFIX; }
	bool	normal()	const		{ return _format==Format::NORMAL;  }
	bool	infix()		const		{ return _format==Format::INFIX;   }
	//bool	prefix()	const		{ return _format==Format::PREFIX;  }

	/* get/set elements */
	int	size()	const			{ return expr.size(); }
	bool	empty() const			{ return expr.size()==0; }
	bool	reserve(const int n)		{ return expr.resize(n); }
	GBody*	operator [] (const int i)	{ return expr[i]; }
const	GBody*	operator [] (const int i) const	{ return expr[i]; }

	/* add elements to expression */
	bool	add(const std::string& token);
	bool	add(GBody* body);
	bool	add(const std::string& token, GBody* body)
				{ return body? add(body) : add(token); }
	GZone&	operator <<(const std::string& token)	{ add(token); return *this; }
	GZone&	operator <<(GBody* body)		{ add(body);  return *this; }
	bool	addPlus( GBody *body);
	bool	addMinus(GBody *body);
	bool	isSubsetOf(GZone& other, const GBody* ignore=nullptr) const;
	bool	merge(GZone& zone);
	dword	hash() const;
	void	setup();

	/* reference links from bodies to current zone */
	void	addReference(GBody* body);
	void	addAllReferences();
	void	removeAllReferences();

	// Generation
	int	generation()		const	{ return _generation; }
	void	generation(int newGen) {
			_generation = newGen;
			//resetBBox();
		}

	// --- bounding box ---
	/** @return 2D bounding box calculated from the segments */
	BBox	bbox2D() const;

	/** @return bounding box of zone, limited to bblimit if any */
const	BBox&	bbox(const BBox* bblimit=nullptr) const {
			if (_bboxState == BBox::State::Invalid) {
				_updateBBox(bblimit);
				_bboxState = BBox::State::Valid;
			}
			return _bbox;
		}
#ifdef OBBOX
	/** @return oriented bounding box of zone */
const	OBBox&	obbox() const {
			if (_obboxState == BBox::State::Invalid) {
				_updateOBBox();
				_obboxState = BBox::State::Valid;
			}
			return _obbox;
		}
#endif
	void	resetBBox() const {
			_bboxState  = BBox::State::Invalid;
#ifdef OBBOX
			_obboxState = BBox::State::Invalid;
#endif
		}

	// --- Operations ---
	/** normal at position (x,y,z) onbody of current zone
	 *
	 * @param x,y,z		position
	 * @param onbody	body to calculate normal on
	 *
	 * @return normal
	 */
	Vector	normal(	const double x, const double y,
			const double z, const GBody* onbody) const;
	Vector	normal(const Point& p, const GBody* onbody) const {
				return normal(p.x, p.y, p.z, onbody);
			}

	/** return surface side on which point (x,y,z) is located
	 *
	 * @param x,y,z		position to return side
	 * @param dsurf		accuracy to consider when we are ON the surface
	 * @param onbody	return body which surface belongs to
	 *
	 * @return side of body Inside,Outside,Surface (@see GBody)
	 */
	GBody::Side inside(const double x, const double y, const double z,
			const double dsurf=SMALL, const GBody** onbody=nullptr) const;

	/** check is the point defined as (x+u*t,y+v*t,z+w*t) towards (u,v,w) at distance t
	 * is inside or outside of current zone
	 * If on boundary the direction is considered outside if the dot product
	 * of the normal and direction is positive
	 *
	 * @param x,y,z	starting position of ray/particle
	 * @param u,v,w	direction of ray/particle
	 * @param t	distance along (u,v,w)
	 *
	 * @return true if inside, false otherwise
	 */
	bool	inside( const double x, const double y, const double z,
			const double u, const double v, const double w) const;
	bool	inside( const Vector& p, const Vector& d) const
			{ return inside(p.x,p.y,p.z, d.x,d.y,d.z); }

	/** cached version of inside. The cache has to be initialized before
	 * the call the max number of bodies.
	 * @warning call the inc() if the position (x,y,z)
	 *          or the direction (u,v,w) has changed
	 * @param x,y,z	starting position of ray/particle
	 * @param u,v,w	direction of ray/particle
	 * @param t	distance along (u,v,w) direction
	 * @param cache	body caching array
	 *
	 * @return true if inside, false otherwise
	 *
	 * @warning body caching has to be initialized (resized) to the max number of bodies
	 *          and the cache.inc() has to be called when ever the position (x,y,z)
	 *          or the direction (u,v,w) has changed.
	 */
	bool	inside(	const double x, const double y, const double z,
			const double u, const double v, const double w,
			const double t, Cache<CBody3D>& cache) const;
	bool	inside(	const Point& p, const Vector& d,
			const double t, Cache<CBody3D>& cache) const
			{ return inside(p.x,p.y,p.z, d.x,d.y,d.z, t, cache); }

	/** return nearest (safe) distance in any direction to exit (or enter) zone
	 * @param x,y,z	starting position
	 *
	 * @return safe distance
	 */
	double	distance(const double x, const double y, const double z) const;

	/** return distance of intersection (tmin) of a ray starting from (x,y,z) with
	 * direction (u,v,w) from in the range (tmin..tmax] distance.
	 * @param x,y,z	starting position
	 * @param u,v,w	direction of ray/particle
	 * @param tmin	I/O starting distance, and on exit contains the returned distance
	 * @param tmax	maximum distance to check
	 * @param toin	true for distance to enter zone, false when we are inside and want to exit
	 * @param cache	body caching array
	 *
	 * @return body the boundary is located on
	 *
	 * @warning body caching has to be initialized (resized) to the max number of bodies
	 *          and the cache.inc() has to be called when ever the position (x,y,z)
	 *          or the direction (u,v,w) has changed.
	 */
	GBody*	distance(const double x, const double y, const double z,
			 const double u, const double v, const double w,
			 double *tmin,   const double tmax,
			 const bool toin, Cache<CBody3D>& cache) const;

	GBody*	distance(const Point& p, const Vector& d,
			 double *tmin,   const double tmax,
			 const bool toin, Cache<CBody3D>& cache) const
			 { return distance(p.x,p.y,p.z, d.x,d.y,d.z, tmin, tmax, toin, cache); }

	GBody*	distance2In(const double x, const double y, const double z,
			    const double u, const double v, const double w,
			    double *tmin, const double tmax,
			 Cache<CBody3D>& cache) const {
				return distance(x,y,z, u,v,w, tmin,tmax, true, cache);
			 }
	GBody*	distance2In(const Point& p, const Vector& d,
			 double *tmin,   const double tmax,
			 Cache<CBody3D>& cache) const
			 { return distance2In(p.x,p.y,p.z, d.x,d.y,d.z, tmin, tmax, cache); }
	GBody*	distance2Out(const double x, const double y, const double z,
			     const double u, const double v, const double w,
			     double *tmin, const double tmax,
			 Cache<CBody3D>& cache) const {
				return distance(x,y,z, u,v,w, tmin,tmax, false, cache);
			 }
	GBody*	distance2Out(const Point& p, const Vector& d,
			 double *tmin,   const double tmax,
			 Cache<CBody3D>& cache) const
			 { return distance2Out(p.x,p.y,p.z, d.x,d.y,d.z, tmin, tmax, cache); }

	// --- CSG operations ---
	void	convertTo(const Format fmt);

	/** verify if an expression is syntactically correct */
	bool	verify() const;

	/** optimize a CSG expression */
	bool	optimize();

	/** optimize possible products inside an rpn expression */
	bool	optimizeProducts(int from=0, int to=-1);

	/** compact zone by removing null pointers */
	bool	compact();

	/** @param from	starting index to search for a zone expression
	 *  @return end-index of a sub-zone inside expression starting from from, -1 on end
	 */
	int	findZone(int from=0);

	/** create a subzone expr[from:to], and convert at the same time
	 * either in PRODUCT or RPN format
	 */
	void	subZone(GZone& subzone, int from, int to);

	/** normalize/expand an rpn expression */
	bool	expand(unsigned limit=0);

	/** @return depth of the rpn expression */
	bool	depth();

	/** @return memory used by Zone */
	size_t	memory() const;

private:
	/** @return index of body inside zone expression */
	int	find(GBody* body) const	{ return expr.find(body); }

	/** @return index of minus term for products, size() on failure */
	int	minusIndex() const {
			assert(std());
			int n = find(&GBody::tminus);
			return (n>=0)? n : size();
		}

	void	_bboxFromPlanes(BBox& bbox) const;
	bool	_insideThreshold(const Vector &v, const Quad *ignore_a,
				const Quad *ignore_b, const Quad *ignore_c) const;
	void	_updateBBox(const BBox* bblimit=nullptr) const;
#ifdef OBBOX
	void	_updateOBBox() const;
#endif

	// rpn helpers
	int	rpnrule(int n, int* start, bool expand=true, bool special=true);
	void	subTerms(int n, int* lowLeft, int* lowRight);
	void	copy(int dst, int src, int length);
static	int	priority(const GBody* body, int* ip=nullptr);

	// Conversion functions
	void	normal2std();
	void	normal2rpn();
	void	rpnorm2normal();

friend class VZone;
friend class CZone;
}; // GZone

std::ostream& operator << (std::ostream&, const GZone&);

#endif
