/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __GEOMETRY_KERNEL_H
#define __GEOMETRY_KERNEL_H

#include <assert.h>
#include <pthread.h>

#include <vector>
#include "tool/stat.h"
#include "graph/viewport.h"
#include "tool/threadpool.h"

#include "gbody.h"
#include "vbody.h"
#include "voxel.h"
#include "engine.h"
#include "gregion.h"

/* ========================== GeometryKernel ========================== */
/** Geometry Kernel class */
class GeometryKernel {
protected:
const   Geometry&	geometry;	/** father geometry object	*/
	GeometryEngine	_engine;	/** main geometry engine	*/
#ifdef THREAD
	GeometryEngine*	threadengines;	/** geometry engine threads	*/
	ThreadPool	threadpool;	/** pool of threads for drawing	*/
#endif

public:
	std::vector<VBody*> bodies;	/** body array			*/
	std::vector<VRegion*> regions;	/** region array		*/
	VVoxel		voxel;		/** voxel structure		*/
	ViewPort	view;		/** Viewing window		*/

protected:
	dword		_bodyShow;	/** ORed body->show flag of all	*/
	double		_minSx2;	/** min Sx2 factor ignore points*/
	int		_errors;	/** errors count, 3D:-1		*/
	dword		_errorhash;	/** error hash variable		*/
	char		_errmsg[128];	/** error message		*/
	IFSTAT(Stat	stat;)		/** Statistics			*/
	pthread_mutex_t	mutex;		/** thread mutex		*/

public:
	GeometryKernel(const Geometry& geo, int nt=-1);
	~GeometryKernel();

	void	clearBodies();
	void	clearRegions();

	/** window */
	void	window(double ex, double ey=0.0);
	void	window(double xmin, double ymin, double xmax, double ymax)
			{ view.window(xmin, ymin, xmax, ymax); }
	void	resize(int w, int h);
	int	width()		const	{ return view.width(); }
	int	height()	const	{ return view.height(); }

	GeometryEngine*	engine()	{ return &_engine; }

	void	initThreads(int nt=0);
#ifdef THREAD
	GeometryEngine*	engine(int i)	{ return &threadengines[i]; }
	int	nthreads()	const	{ return threadpool.nthreads(); }
#else
	int	nthreads()	const	{ return 0; }
#endif

	/* bodies */
	size_t	nBodies()		const	{ return geometry.nBodies(); }
	// XXX WARNING unprotected for limits
	VBody*	getBody(const size_t id)	{ return bodies[id]; }
	VBody*	getBody(const std::string& name) {
			const GBody* body = geometry.getBody(name);
			return body?getBody(body->id()):NULL;
		}
	VBody*	getBody(const GBody* body)	{ return bodies[body->id()]; }
	void	clearVisibleBodies();
	void	updateBodyShow();
	dword	bodyShow()	const		{ return _bodyShow; }

	/* projection bodies */
	void	projectBodyClear();
	void	projectBodyAdd(int idx);
	int	projectBody(int i)	const	{ return _engine.projectBody(i); }
	int	projectBodyCount()	const	{ return _engine.projectBodyCount(); }

	/* clipping bodies */
	void	clipBodyClear();
	void	clipBodyAdd(int idx);
	void	clipBodyNegative(int idx);
	int	clipBody(int i)		const	{ return _engine.clipBody(i); }
	int	clipBodyCount()		const	{ return _engine.clipBodyCount(); }

	/* regions */
	VRegion* getRegion(const size_t id) {
			if (id<regions.size())
				return regions[id];
			else
				return NULL;
		}
	VRegion* getRegion(const std::string& name) {
			const GRegion* region = geometry.getRegion(name);
			return region? getRegion(region->id()) : NULL;
		}
	VRegion* getRegion(GRegion*region)	{ return regions[region->id()]; }
	size_t	 nRegions()		const	{ return geometry.nRegions(); }

	/* transformations */
	void	matrix(const Matrix4& m);
	void	origin(const double x, const double y, const double z);
	void	origin(double *x, double *y, double *z)	{ view.origin(x,y,z); }
	void	moveViewOriginTo0();

	/* errors */
	int	errors()	const	{ return _errors; }
	void	clearErrors()		{ _errors = 0; }
	void	clearErrorHash()	{ _errorhash = 0; }
	dword	errorHash()	const	{ return _errorhash; }
	void	clearError3D()		{ if (_errors<0) _errors = 0; }
	void	setError3D()		{ if (!_errors) _errors = -1; }
	VBody*	error(const int err, double* u, double* v,
			ZoneOfPoint* pIn, ZoneOfPoint* pOut,
			double* umin, double* umax,
			double* vmin, double* vmax);

	/* error message */
	void	errorReset()		{ _errmsg[0] = 0; }
	void	error(const char *msg)	{ if (_errmsg[0]==0) strcpy(_errmsg, msg); }
const	char*	error()		const	{ return _errmsg; }

	/* body operations */
	void	makeBodyConics(bool all);
	void	intersectBody(VBody* a, bool all);
	bool	scanDirection(VBody* body, const int cid,
			const double u, const double v,
			ZoneOfPoint* pIn, ZoneOfPoint* pOut);
	void	scanBodySegments(bool all);
	void	scanBodySegments(VBody* body, GeometryEngine* eng, bool all);

	/* region operations (MOVE TO KERNEL) */
	void	updateRegionLocation(bool all);
	bool	correctOverlaps(VRegion *region, int zoneid=-1);
	VBody*	zoneOverlaps(VZone* zone, int *side);
	int	bodySide(VZone* zone, VBody* body);
	int	segmentSide(VZone* zone, VBody* body, int cid, int from, int to);
	bool	isZoneExist(VZone* zone, VBody* body);

	/* threads */
	void	lock() {
			pthread_mutex_lock(&mutex);
			geometry.lockRead();
		}
	void	unlock() {
			geometry.unlockRead();
			pthread_mutex_unlock(&mutex);
		}
	void	stopThread()	{
#ifdef THREAD
			threadpool.stop();
#endif
		}
	bool	isRunning() {
#ifdef THREAD
			return threadpool.isRunning();
#else
			return false;
#endif
		}
#ifdef STAT
	void	resetStatistics();
	void	mergeStatistics();
#endif

	/** Derive */
	void	derive();

	/* Info */
	size_t	bodiesMemory()	const;
	size_t	regionsMemory()	const;
	size_t	memory()	const;
	void	printMemory()	const;

friend class Layer;
friend class D2Layer;
friend class D3Layer;
}; // GeometryKernel
#endif
