/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	1-Jun-2016
 */

#ifndef __GEOMETRY_MESH_H
#define __GEOMETRY_MESH_H

#include <algorithm>

#include "edge.h"
#include "face.h"

#include "math/bbox.h"
#include "tool/list.h"
#include "tool/array.h"

class GBody;
class Vector;
class Matrix4;

// =================================== Mesh ====================================
class Mesh {
	using BodyList = List<GBody*>;
private:
	enum class FlagType {
			AllFaces,
			PositivePid,
			NegativePid,
			EqualPid
	};

private:
	VertexArray	_vertices;	/** array of vertices	*/	// FIXME be careful on resize->reserve()
	EdgeArray	_edges;		/** array of edges	*/	// FIXME be careful on resize->reserve()
	FaceVector	_faces;		/** array of faces	*/
	BBox		_bbox;		/** bounding box	*/
private:
	double		eps;		/* precision for operations */

public:
	/** constructor*/
	Mesh();

	/** copy constructor */
	Mesh(const Mesh& m);

	/** destructor */
	~Mesh()	{ free(); };

	/** clean up mesh */
	void	free();
	void	freeVertices();
	void	freeEdges();
	void	freeFaces();
	void	reserve(size_t nv, size_t ne, size_t nf);
	void	allocateVertices(size_t nv);
	bool	isEmpty()		const	{ return nvertices() == 0; }

	// --- Vertex operations ---
	size_t	nvertices()		const	{ return _vertices.size(); }
	Vertex&	vertex(size_t i)		{ return *_vertices[i]; }
const	Vertex&	vertex(size_t i)	const	{ return *_vertices[i]; }
	Vertex*	add(const Vertex& v);
	Vertex*	add(Vertex* v);
	int	findVertex(const Vertex&) const;
//	int	vertexIndex(Vertex*)	const	{ return _vertices.find(v); }

	// --- Edge operations ---
	size_t	nedges()		const	{ return _edges.size(); }
	Edge*	edge(size_t i)			{ return _edges[i]; }
const	Edge*	edge(size_t i)		const	{ return _edges[i]; }
	Edge*	add(Vertex* a, Vertex* b, bool show=true);
	Edge*	add(size_t a, size_t b, bool show=true)
			{ return add(_vertices[a], _vertices[b], show); }

	/** Find an edge that contains the two vertices
	 * @param a first vertex of the edge
	 * @param b second vertex of the edge
	 * @return pointer to the edge, NULL otherwise */
	int	findEdge(Vertex* a, Vertex* b) const {
			Edge e(a,b);
			return _edges.find(&e);
//			// FIXME to be checked
//			auto it = std::find(_edges.begin(), _edges.end(), e);
//			if (it == _edges.end())
//				return -1;
//			else
//				return it - _edges.begin();
		}

	// --- Face operations ---
	size_t	nfaces()		const	{ return _faces.size(); }
	Face*	face(size_t i)			{ return _faces[i]; }
const	Face*	face(size_t i)		const	{ return _faces[i]; }
//	Face*	add(Face* f)			{ _faces.add(f); return f; }
	Face*	add(Face* f)			{ _faces.push_back(f); return f; }
	Face*	add(Edge* AB, Edge* BC, Edge* CA){ return add(new Face(AB, BC, CA)); }
	Face*	add(Vertex* A, Vertex* B, Vertex* C, bool ab=true, bool bc=true, bool ca=true);
	Face*	add(size_t a, size_t b, size_t c, bool ab=true, bool bc=true, bool ca=true)
				{ return add(_vertices[a], _vertices[b], _vertices[c], ab, bc, ca); }

	void	forEachVertex(VertexFunc, void* arg=NULL);
	void	forEachEdge(EdgeFunc, void* arg=NULL);
	void	forEachFace(FaceFunc, void* arg=NULL);

	// bounding box
const	BBox&	bbox()			const	{ return _bbox; }
	void	calcBbox();

	// creation
	void	createParallelepiped(const Vertex& P, const Vector& X,
				const Vector& Y, const Vector& Z);

	// bodies fitting
	void	fit(BodyList& list, const double ang);

	// operations
	bool	isClosed();
	bool	isOrientable();
	void	makeOrientable();
	int	problematicEdges() const;
	double	surface() const;
	double	volume() const;

	void	flip();
	void	process();
	void	transform(const Matrix4& matrix);

	size_t	memory() const;

protected:
	int	fitPlanes(BodyList& list, size_t minFaces, int& pid);
//	int	fitCylinders(BodyList& list, size_t minFaces, int& pid);

	void	clearProcessedFlag(int b=0, FlagType type=FlagType::AllFaces);
	bool	addNeighborsWithSimilarNormal(Face* f, FaceVector& fara, const double dot, int pid);
	bool	addNeighborsCylinder(Face* f, FaceVector& fara,
				bool& first, double* dot, Vector* axis,
				int pid, double acc);
	void	faces2Vertices(FaceVector& fara, VertexArray& pts) const;

public:
	/** print mesh */
friend	std::ostream& operator << (std::ostream&, Mesh& mesh);
}; // Mesh

std::ostream& operator << (std::ostream& s, Mesh& mesh);

#endif
