/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	10-Oct-2010
 */

#include "torusbody.h"

#include "mesh.h"

#define N_TORUS_LON	16
#define N_TORUS_LAT	32

// ================================ GTorusBody =================================
/** set */
void GTorusBody::set(const double x, const double y, const double z,
		     const double a, const double b, const double c)
{
	P.set(x, y, z);
	if (a<TOOSMALL or b<TOOSMALL or c<TOOSMALL)
		throw std::string("Too small or negative radius");
	ra = a;
	rb = b;
	rc = c;

	switch (type()) {
		case Type::TRX:
			Z = Vector::Xo;
			X = Vector::Yo;
			Y = Vector::Zo;
			break;

		case Type::TRY:
			Z = Vector::Yo;
			X = Vector::Zo;
			Y = Vector::Xo;
			break;

		case Type::TRZ:
			Z = Vector::Zo;
			X = Vector::Xo;
			Y = Vector::Yo;
			break;
		default:
			assert(false);
	}
} // set

/** set */
void GTorusBody::set(double *what)
{
	resetBBox();
	round0(what, nwhats(), SMALL2);
	set(what[0], what[1], what[2],
	    what[3], what[4], what[5]);
} // set

/* --- get --- */
int GTorusBody::get(double *what) const
{
	what[0] = P.x;
	what[1] = P.y;
	what[2] = P.z;
	what[3] = ra;
	what[4] = rb;
	what[5] = rc;
	return 6;
} // get

/** @return bounding box of body */
void GTorusBody::updateBBox() const
{
	Point p = GBody::position();
	_bbox.add(Point(p - (ra+rc)*(vectorX()+vectorY()) - rb*vectorZ()));
	_bbox.add(Point(p + (ra+rc)*(vectorX()+vectorY()) + rb*vectorZ()));
} // updateBBox

#ifdef OBBOX
/** @return bounding box of body */
void GTorusBody::updateOBB(bool /*in*/) const
{
	assert(false);
	//if (!in) {
	//	Point p = GBody::position();
	//	_obbox.add(Point(p - (ra+rc)*(vectorX()+vectorY()) - rb*vectorZ()));
	//	_obbox.add(Point(p + (ra+rc)*(vectorX()+vectorY()) + rb*vectorZ()));
	//}
} // updateOBB
#endif

/** createQuads */
void GTorusBody::createQuads()
{
	GBody::createQuads();
	// XXX XXX XXX XXX XXX XXX
	// FIXME FOR TESTING create two cylinders and two planes
	// XXX XXX XXX XXX XXX XXX
	//
	// Create the cylinder around Z then transform
	//  Cxx Cyy Czz  Cxy  Cxz  Cyz  Cx   Cy   Cz    C
	addQuad(-1.0,-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, Sq(ra-rc));
	addQuad( 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,-Sq(ra+rc));

	Matrix4 M, Minv, T;
	Minv.make(X,Y,Z);
	Minv.inverse();
	T.translate(P);
	M.multiply(T, Minv);
	Minv.inverse(M);

	// transform first quad
	_Q[0].transform(Minv, M);
	_Q[0].normalize();

	_Q[1].transform(Minv, M);
	_Q[1].normalize();

	// two planes for the caps
	//        Cx     Cy     Cz      C
	addQuad(-Z.x,-Z.y,-Z.z,  Z*P - rb);	// base
	addQuad( Z.x, Z.y, Z.z,-(Z*P + rb));	// apex
} // createQuads

/* createMesh */
void GTorusBody::createMesh()
{
	GBody::createMesh();
	bool first = _mesh->isEmpty();
	if (first) _mesh->allocateVertices(N_TORUS_LAT * N_TORUS_LON);

	const double stepf = PI2 / (double)N_TORUS_LAT;
	const double step0 = PI2 / (double)N_TORUS_LON;

	int k = 0;
	double angf = 0.0;
	for (int j=0; j<N_TORUS_LAT; j++, angf += stepf) {
		double sf, cf;
		bsincos(angf, &sf, &cf);
		double ang0 = 0.0;
		for (int i=0; i<N_TORUS_LON; i++, ang0 += step0) {
			double s0, c0;
			bsincos(ang0, &s0, &c0);
			_mesh->vertex(k++) = P +
					(ra + rc*c0)*cf * X +
					(ra + rc*c0)*sf * Y +
					rb*s0*Z;
		}
	}

	_mesh->calcBbox();

	if (first) {
		k = 0;
		for (int j=0; j<N_TORUS_LAT; j++) {
			for (int i=0; i<N_TORUS_LON; i++, k++) {
				int n = (j<N_TORUS_LAT-1)? k+N_TORUS_LON : i;

				int m = (i<N_TORUS_LON-1)? n+1 : n+1-N_TORUS_LON;
				_mesh->add(k, n, m, true, true, false);

				n = (i<N_TORUS_LON-1)? k+1 : k+1-N_TORUS_LON;
				_mesh->add(k, m, n, false, true, true);
			}
		}

		_mesh->process();
		assert(_mesh->isClosed());
		assert(_mesh->isOrientable());
//#if _DEBUG>2
		std::cout << std::endl;
		std::cout << "TRXYZ Mesh ";
		std::cout << " isClosed=" << _mesh->isClosed();
		std::cout << " isOrientable=" << _mesh->isOrientable() << std::endl;
		std::cout << "TRZ volume=" << _mesh->volume() << std::endl;
//#endif
	}
} // createMesh
