/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2011
 */

#ifndef __GEOMETRY_VREGION_H
#define __GEOMETRY_VREGION_H

#include "geo.h"
#include "gregion.h"

#include "tool/array.h"
#include "math/vector.h"

class VZone;
class GeometryKernel;
class GeometryEngine;

/** Viewport Region class */
class VRegion {
private:
	GRegion*	_region;	// region reference
	Array<VZone*>	_zones;		// List of zones
	GeometryKernel*	_kernel;	// Kernel where region belongs

public:
	char	label[16];	// label or value to display

private:
	int	_generation;	// Generation number, to be checked for validity
				// against the _region generation
	dword   _hash;
	bool	_location;	// if true viewport location vs region is known (given by in)
	VZone*	_in;		// if viewport is inside or outside the region
	dword	_color;		// region color
	double	_value;		// region color value
	byte	_alpha;		// Transparency level [0..255]

public:
	VRegion(GRegion* reg=NULL, GeometryKernel* k=NULL) :
		_color(0xA0A0A0),
		_value(0.0),
		_alpha(0)
		{ label[0] = 0; init(reg, k); }
virtual	~VRegion()			{ clear(); _region=NULL; }

	void	init(GRegion*reg=NULL, GeometryKernel* k=NULL);
	void	clear();

	GRegion* region()	const	{ return _region; }
const	std::string&	name()	const	{ return _region->name(); }
const	char*	cname()		const	{ return _region->cname(); }
	GRegion::Type type()	const	{ return _region->type(); }
	int	id()		const	{ return _region->id(); }
	dword	hash()		const	{ return _hash; }

	int	show()		const	{ return _region->show; }

	void	color(dword c)		{ _color = c & FLAG_COLORMASK; }
	dword	color()		const	{ return _color; }

	void	value(double v)		{ _value = v; }
	double	value()		const	{ return _value; }

	void	alpha(byte a)		{ _alpha = a;     }
	byte	alpha()		const	{ return _alpha;      }
	bool	transparent()	const	{ return _alpha==255; }
	bool	notTransparent()const	{ return _alpha==0;   }

	// Generation/valid/invalid
	bool	valid()		const	{ return _generation == _region->generation(); }
	bool	invalid()	const	{ return _generation != _region->generation(); }
	void	setInvalid()		{ _generation = -1;  }
	void	setValid()		{ _generation = _region->generation(); }

	int	size()		const	{ return _zones.size(); }
const	Array<VZone*>& zones()	const	{ return _zones; }
	VZone*	zone(int i)	const	{ return _zones[i]; }

	bool	ignore()	const	{ return _location && !_in; }

	void	updateLocation();
	VZone*	inside2D(GeometryEngine* engine,
			 const double  x, const double  y, const double  z,
			 const double dx, const double dy, const double dz) const;
	VZone*	inside2D(GeometryEngine* engine,
			 const Vector& p, const Vector& d) const
			{ return inside2D(engine, p.x,p.y,p.z, d.x,d.y,d.z); }
	VZone*	inside(GeometryEngine* engine,
			 const double  x, const double  y, const double  z,
			 const double dx, const double dy, const double dz) const;
	VZone*	inside(GeometryEngine* engine,
			 const Vector& p, const Vector& d) const
			{ return inside(engine, p.x,p.y,p.z, d.x,d.y,d.z); }
	VZone*	inside(GeometryEngine* engine,
			  const double  x, const double  y, const double  z,
			  const double dx, const double dy, const double dz,
			  const double t ) const;
	VZone*	inside(GeometryEngine* engine,
			  const Vector& p, const Vector &d, const double t) const
			{ return inside(engine, p.x,p.y,p.z, d.x,d.y,d.z, t); }

	void	showBodies(int show);

	size_t	memory() const;

friend	class	VZone;
}; // VRegion

std::ostream& operator << (std::ostream&, const VRegion&);

#endif
