/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	02-Feb-2015
 */

#include "histogram.h"
#include "math/bmath.h"

#include <math.h>
#include <string.h>

#include <iostream>

/** Histogram */
Histogram::Histogram(const char *atitle, int bins, double low, double high)
	: xbins(0), xlow(0.0), xhigh(1.0), h(NULL), eh(NULL)
{
	if (atitle) title(atitle);
	set(bins, low, high);
} // Histogram

/** Histogram */
Histogram::Histogram(const Histogram& hist)
	: xbins(0), h(NULL), eh(NULL)
{
	if (not hist.title().empty()) title(hist.title());
	set(hist.xbins, hist.xlow, hist.xhigh);
	memcpy(h, hist.h, sizeof(double)*xbins);
	memcpy(eh, hist.eh, sizeof(double)*xbins);

	_entries = hist._entries;
	_under   = hist._under;
	_over    = hist._over;
	_sumw    = hist._sumw;
	_sumw2   = hist._sumw2;
	_sumwx   = hist._sumwx;
	_sumwx2  = hist._sumwx2;
} // Histogram

/** ~Histogram */
Histogram::~Histogram()
{
	free();
} // ~Histogram

/** free */
void Histogram::free()
{
	if (h)  delete [] h;
	if (eh) delete [] eh;
	h  = NULL;
	eh = NULL;
	xbins = 0;
} // free

/** set */
void Histogram::set(int n, double low, double high)
{
	if (n != xbins) {
		free();
		xbins = n;
	}
	setLimits(low,high);

	if (xbins && h==NULL) {
		h  = new double[xbins];
		eh = new double[xbins];
	}
	clear();
} // set

/** setLimits */
void Histogram::setLimits(double low, double high)
{
	xlow  = low;
	xhigh = high;
	if (xbins) xstep = (xhigh-xlow) / (double)xbins;
} // setLimits

/** clear */
void Histogram::clear()
{
	_entries = 0;
	_under   = 0.0;
	_over    = 0.0;
	_sumw    = 0.0;
	_sumw2   = 0.0;
	_sumwx   = 0.0;
	_sumwx2  = 0.0;
	memset(h,  0, sizeof(h[0]) *xbins);
	memset(eh, 0, sizeof(eh[0])*xbins);
} // clear

/** set */
void Histogram::set(int bin, double content)
{
	_entries++;
	if (bin<0) {
		_sumw  -= _under;
		_under  = content;
		_sumw  += content;
	} else
	if (bin>=xbins) {
		_sumw  -= _over;
		_over   = content;
		_sumw  += content;
	} else {
		_sumw  -= h[bin];
		h[bin]  = content;
		_sumw  += content;
	}
} // set

/** fill histogram */
int Histogram::fill(const double x, const double w)
{
	_entries++;
	_sumw   += w;
	_sumw2  += w*w;
	_sumwx  += w*x;
	_sumwx2 += w*x*x;
	if (x<xlow) {
		_under += w;
		return -1;
	} else
	if (x>xhigh) {
		_over  += w;
		return -1;
	} else {
		int i = geti(x);
		h[i]  += w;
		eh[i] += w*w;		// during filling the eh holds the sum of w^2
		return i;
	}
} // fill

/** normalize histogram with f value
 * @param f normalization factor
 */
void Histogram::norm(const double f)
{
	assert(xbins);
	_under  *= f;
	_over   *= f;
	_sumw   *= f;
	_sumw2  *= f*f;
	_sumwx  *= f;
	_sumwx2 *= f;
	for (int i=0; i<xbins; i++) {
		h[i]  *= f;
		eh[i] *= f;
	}
} // norm

/** convert the histogram to cumulative one loop: h[i+1] += h[i]
 *  ignoring the _over and _under!
 * @param reverse from max to 0.0
 */
void Histogram::cumulative(bool reverse)
{
	assert(xbins);
	if (!reverse) {
		h[0] += _under;
		for (int i=1; i<xbins; i++)
			h[i] += h[i-1];
			// errors?
		_over += h[xbins-1];
	} else {
		h[xbins-1] += _over;
		for (int i=xbins-2; i>=0; i--)
			h[i] += h[i+1];
			// errors?
		_under += h[0];
	}
} // cumulative

/** convert a log10(histogram) to isolethargic */
void Histogram::isolethargic()
{
	norm(exp10(xstep/2.0) / exp10(xstep-1.0));
} // isolethargic

/** save */
void Histogram::save(FILE *fout, int columns, bool lastdup)
{
	fprintf(fout,"# title   %s\n",    ctitle());
	fprintf(fout,"# xmin    %.10g\n", xlow);
	fprintf(fout,"# xmax    %.10g\n", xhigh);
	fprintf(fout,"# xbins   %d\n"   , xbins);
	fprintf(fout,"# xstep   %.10g\n", xstep);
	fprintf(fout,"# n       %d\n"   , _entries);
	fprintf(fout,"# under   %.10g\n", _under);
	fprintf(fout,"# over    %.10g\n", _over);
	fprintf(fout,"# sumw    %.10g\n", _sumw);
	fprintf(fout,"# sumw2   %.10g\n", _sumw2);
	fprintf(fout,"# sumwx   %.10g\n", _sumwx);
	fprintf(fout,"# sumwx2  %.10g\n", _sumwx2);

	double x = xlow;
	for (int i=0; i<xbins; i++) {
		double xh = x + xstep;
		switch (columns) {
			case 3:
				fprintf(fout,"%.10g %.10g %.10g\n",x,h[i],eh[i]);
				break;

			case 4:
				fprintf(fout,"%.10g %.10g %.10g %.10g\n",x,xh,h[i],eh[i]);
				break;

			default:	/* 2 */
				fprintf(fout,"%.10g %.10g\n",x,h[i]);
		}
		x = xh;
	}

	/* correction for gnuplot. Duplicate the last line */
	if (lastdup) {
		int i = xbins-1;
		switch (columns) {
			case 3:
				fprintf(fout,"%.10g %.10g %.10g\n",x,h[i],eh[i]);
				break;

			case 4:
				fprintf(fout,"%.10g %.10g %.10g %.10g\n",x,x,h[i],eh[i]);
				break;

			default:	/* 2 */
				fprintf(fout,"%.10g %.10g\n",x,h[i]);
		}
	}
} // save

/** load */
bool Histogram::load(FILE *)
{
	assert(0);
	return true;
} // load
