/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __MATH_QUAD_H
#define __MATH_QUAD_H

#include <ostream>

#include "eps.h"
#include "point.h"

class BBox;
#ifdef OBBOX
class OBBox;
#endif
class Matrix3;
class Matrix4;

/**
 * Generalized quadratic surface
 * Q(x,y,z) =   Cxx*x^2 + Cyy*y^2 + Czz*z^2
 *            + Cxy*xy  + Cxz*xz  + Cyz*yz
 *            + Cx*x    + Cy*y    + Cz*z    + C  = 0
 *
 * Represented in matrix format
 *          0       1      2      3
 *     0 / Cxx     Cxy/2  Cxz/2  Cx/2 \
 *     1 | Cxy/2   Cyy    Cyz/2  Cy/2 |
 *     2 | Cxz/2   Cyz/2  Czz    Cz/2 |
 *     3 \ Cx/2    Cy/2   Cz/2   C    /
 *
 * @see doc/quad_surfaces_types.pdf
 * Type of surface
 *   #  Rank(e)  Rank(E)   D   signs-k  Type
 *  --  -------  ------- ----- -------  ----------------------------
 *   1     3        4      <0    same   Real Ellipsoid
 *   2     3        4      >0    same   Img Ellipsoid
 *
 *   3     3        4      >0    diff   Hyperboloid of 1 sheet
 *   4     3        4      <0    diff   Hyperboloid of 2 sheets
 *
 *   5     3        3            diff   Real Quadric Cone
 *   6     3        3            same   Img Quadric Cone
 *
 *   7     2        4      <0    same   Elliptic Paraboloid
 *   8     2        4      >0    diff   Hyperbolic Paraboloid
 *
 *   9     2        3            same   Real Elliptic Cylinder
 *  10     2        3            same   Img Elliptic Cylinder
 *
 *  11     2        3            diff   Hyperbolic Cylinder
 *
 *  12     2        2            diff   Real Intersecting Planes
 *  13     2        2            same   Img Intersecting Planes
 *
 *  14     1        3                   Parabolic Cylinder
 *
 *  15     1        2                   Real Parallel Planes
 *  16     1        2                   Img Parallel Planes
 *
 *  17     1        1                   Coincident Planes
 *
 *  Q(x,y,z) =  Ax^2 +  By^2 + C^z^2 +
 *             2Fyz  + 2Gzx  + 2Hxy  +
 *             2Px   + 2Qy   + 2Rz   + D = 0
 *
 *     / A  H  G \        / A  H  Q  P \
 * e = | H  B  F |    E = | H  B  F  Q |
 *     \ G  F  C /        | G  F  C  R |
 *                        \ P  Q  R  D /
 *
 * k = eigenvalues of e-matrix
 *      / A-k   H    G  \
 *      |  H   B-k   F  | = 0
 *      \  G    F   C-k /
 */
class Quad {
public:
	// Quadratic types that can be handled
	enum class Type : int {
		PLANE,			// planes should be first
			PLANE_X,	// important for isplane()
			PLANE_Y,
			PLANE_Z,	// subtypes of plane

		QUAD,			// Generic quad

		INTERSECTING_PLANES,
		PARALLEL_PLANES,
		COINCIDENT_PLANES,

		SPHERE,
		ELLIPSOID,

		CYLINDER,
			CYLINDER_X,	// subtypes of cylinder
			CYLINDER_Y,
			CYLINDER_Z,

		ELLIPTIC_CYLINDER,
			ELLIPTIC_CYLINDER_X,
			ELLIPTIC_CYLINDER_Y,
			ELLIPTIC_CYLINDER_Z,

		CONE,			// One side cone
			CONE_X,		// subtypes of cone
			CONE_Y,
			CONE_Z,

		HYPERBOLOID_1_SHEET,
		HYPERBOLOID_2_SHEET,
		ELLIPTIC_PARABOLOID,
		HYPERBOLIC_PARABOLOID,
		HYPERBOLIC_CYLINDER,
		PARABOLIC_CYLINDER,

		IMG_INTERSECTING_PLANES,
		IMG_PARALLEL_PLANES,
		IMG_ELLIPSOID,
		IMG_CONE,
		IMG_ELLIPTIC_CYLINDER,

		DEGENERATE
	};

public:
	double	Cxx, Cyy, Czz;		// Quad coefficients
	double	Cxy, Cxz, Cyz;
	double	Cx,  Cy,  Cz;
	double	C;

public:
	Type	_type;			// all high orders are zero
//	union	{
//		struct			// depending type set parameters
	Point	P;			// center position of quad (related to Cx,Cy,Cz terms
	Vector	Z;			// local axis system
	double	R;			// radius of sphere, cylinder
	double	Ra;			// apex radius
	double	slope;			// slope for cone (Ra-R)/H
	double	s21;			// slope²+1
//	}

	double	aCxx, aCyy, aCzz;	// absolute value of un-normalized coefficients
	double	aCxy, aCxz, aCyz;	// for accuracy calculation
	double	aCxyz0;

	double	_dist1D;		// caching variables calculated in init()
	double	_dist2a;		// used for distance

static	const char* _typeStr[];

public:
	Quad() : Cxx(0.0), Cyy(0.0), Czz(0.0),
		 Cxy(0.0), Cxz(0.0), Cyz(0.0),
		 Cx(0.0),  Cy(0.0),  Cz(0.0), C(0.0), _type(Type::DEGENERATE) {}

	Quad(const double cx,  const double cy,  const double cz,
	     const double c);

	Quad(const double cxx, const double cyy, const double czz,
	     const double cxy, const double cxz, const double cyz,
	     const double cx,  const double cy,  const double cz,
	     const double c);

	Quad(Type t, const Point& p, const double v);
	Quad(Type t, const Point& p, const Vector& z, const double v);
	Quad(Type t, const Point& p, const Vector& z, const double v, const double v2);

	// copy & assignment operators
	Quad(const Quad& q) = default;
	Quad& operator =(const Quad&) = default;
	Quad& operator =(const Matrix4&);

private:
	// set quad from parameters
	void	reset();
	void	set(const double cx,  const double cy,  const double cz,
		    const double c);
	void	set(const double cxx, const double cyy, const double czz,
		    const double cxy, const double cxz, const double cyz,
		    const double cx,  const double cy,  const double cz,
		    const double c);
	void	set(Type t, const Point& aP, const double value);
	void	set(Type t, const Point& aP, const Vector& aZ,
			const double value);
	void	set(Type t, const Point& aP, const Vector& aZ,
			const double value, const double value2);
public:

	// type handling
	Type	type()		const	{ return _type; }
const	char*	typeStr()	const	{ return _typeStr[(int)_type]; }
static	const char* typeStr(Type t)	{ return _typeStr[(int)t]; }

	Type	detectType()	const;

	bool	isplane()	const	{ return _type<=Type::PLANE_Z; }

	/** evaluate at x,y with z=0 */
	double	operator ()(const double x, const double y) const {
			if (isplane())
				return Cx*x + Cy*y + C;
			else
				return (Cxx*x + Cxy*y + Cx)*x + (Cyy*y + Cy)*y + C;
		}

	/** evaluate at x,y,z */
	double	operator ()(const double x, const double y, const double z) const {
			if (isplane())
				return Cx*x + Cy*y + Cz*z + C;
			else
				return	(Cxx*x + Cxy*y + Cxz*z + Cx)*x +
					(Cyy*y + Cyz*z + Cy)*y +
					(Czz*z + Cz)*z + C;
		}

	double	operator ()(const Point& p) const { return (*this)(p.x, p.y, p.z); }

	double	Q(const double x, const double y) const			{ return (*this)(x,y); }
	double	Q(const double x, const double y, const double z) const	{ return (*this)(x,y,z); }
	double	Q(const Point& p) const					{ return (*this)(p); }

	Quad&	operator *= (const double d);

	/** evaluate accuracy at x,y with z=0 */
	double	acc(const double x, const double y, const double eps) const {
			if (isplane())
				return eps*(Abs(x) + Abs(y) + aCxyz0);
			else {
				double ax = Abs(x);
				double ay = Abs(y);
				return eps*( x*x + y*y +
					     /*2.0*(*/aCxx*ax + aCyy*ay/*)*/ +
					     ax*ay +
					     aCxy*(ax + ay) +
					     ax + ay +
					     aCxyz0);
			}
		}

	/** evaluate accuracy at x,y,z */
	double	acc(const double x, const double y, const double z, const double eps) const {
			if (isplane())
				return eps*(Abs(x) + Abs(y) + Abs(z) + aCxyz0);
			else {
				double ax = Abs(x);
				double ay = Abs(y);
				double az = Abs(z);
				return eps*( x*x + y*y + z*z +
					     /*2.0*(*/aCxx*ax + aCyy*ay + aCzz*az/*)*/ +
					     ax*ay + ax*az + ay*az +
					     aCxy*(ax + ay) +
					     aCxz*(ax + az) +
					     aCyz*(ay + az) +
					     ax + ay + az +
					     aCxyz0);
			}
		}

	/** evaluate gradient at x,y with z=0 */
	void	grad(const double x, const double y,
		     double* gx, double* gy, double* gz) const {
			if (isplane()) {
				*gx = Cx;
				*gy = Cy;
				*gz = Cz;
			} else {
				*gx = 2.0*Cxx*x + Cxy*y + Cx;
				*gy = 2.0*Cyy*y + Cxy*x + Cy;
				*gz = Cxz*x + Cyz*y + Cz;
			}
		}
	/** evaluate gradient at x,y,z */
	void	grad(const double x, const double y, const double z,
		     double* gx, double* gy, double* gz) const {
			if (isplane()) {
				*gx = Cx;
				*gy = Cy;
				*gz = Cz;
			} else {
				*gx = 2.0*Cxx*x + Cxy*y + Cxz*z + Cx;
				*gy = 2.0*Cyy*y + Cxy*x + Cyz*z + Cy;
				*gz = 2.0*Czz*z + Cxz*x + Cyz*y + Cz;
			}
		}
	Vector	grad(const Point& r) const {
			double gx, gy, gz;
			grad(r.x,r.y,r.z, &gx,&gy,&gz);
			return Vector(gx,gy,gz);
		}

	double	distance(const double x, const double y, const double z) const;
	double	distance(const Point& p) const	{ return distance(p.x, p.y, p.z); }
	int	distance(const double x, const double y, const double z,
			 const double u, const double v, const double w,
			 double* tmin, double* tmax) const;
	int	distance(const Point& p, const Vector& v,
			 double* tmin, double* tmax) const
				{ return distance(p.x,p.y,p.z, v.x,v.y,v.z, tmin, tmax); }
	void	translate(const double dx, const double dy, const double dz);
	void	translate(const Vector& v)	{ translate(v.x, v.y, v.z); }
	void	matrix3(Matrix3* m) const;
	void	matrix(Matrix4* M) const;
	void	transform(const Matrix4& M, const Matrix4& Minv);
	void	negate();
	void	normalize();

	// bounding box
	BBox	bbox() const;
#ifdef OBBOX
	OBBox	obbox() const;
#endif
private:
	void	computeAbs(const double norm=1.0);
	void	findSubType(double eps=TOOSMALL);
	void	init();
}; // Quad

std::ostream& operator << (std::ostream&, const Quad&);

#endif
