/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	28-Oct-2019
 */

#ifndef __MATH_QUATERNION_H
#define __MATH_QUATERNION_H

#include <math.h>

#include "vector.h"
#include "matrix3.h"
#include "matrix4.h"

// ================================ Quaternion =================================
/**
 * Quaternions class  s+V = s + V.x*i + V.y*j + V.z*k
 *
 * Unit quaternions can represent rotation along the vector V for an
 * angle theta()
 * q = [cos(Θ/2), u*sin(Θ/2), v*sin(Θ/2), w*sin(Θ/2)]
 *
 * Rotation matrix: if v=(l,m,n)
 *    /x'\   / l²(1-cosΘ)+ cosΘ     lm(1-cosΘ)-nsinΘ    nl(1-cosΘ)+msinΘ \ /x\
 *    |y'| = | lm(1-cosΘ)+nsinΘ     m²(1-cosΘ)+ cosΘ    mn(1-cosΘ)-lsinΘ | |y|
 *    \z'/   \ nl(1-cosΘ)-msinΘ     mn(1-cosΘ)+lcosΘ    n²(1-cosΘ)+ cosΘ / \z/
 *
 * using the quaternions
 *      w=[0,x',y',z']
 *      v=[0,x, y, z]
 *      q = [cos(Θ/2), l*sin(Θ/2), m*sin(Θ/2), n*sin(Θ/2))]
 *
 *  w = q v q*    (Note that q^-1 = q* if |q|=1)
 */
class Quaternion {
public:
	double	s;	// scalar value
	Vector	V;	// vector value
public:
	Quaternion() : s(0) {}

	/** Constructor for an arbitrary quaternion with scalar and vector */
	Quaternion(double a, double x, double y, double z) : s(a), V(x,y,z) {}

	/** Constructor for an arbitrary quaternion with scalar and vector */
	Quaternion(double a, const Vector& v) : s(a), V(v) {}

	/** Constructor for a rotation quaternion with an axis and rotation angle */
	Quaternion(const Vector& v, const double angle) {
			s = cos(angle/2.0);
			V = v.unit()*sin(angle/2.0);
		}

	/** Constructor for a rotation quaternion with an rotation matrix */
	Quaternion(const Matrix4& M);

//virtual ~Quaternion() {}
	Quaternion(const Quaternion& q) = default;
	Quaternion& operator = (const Quaternion& q) = default;

	/** @return components by index */
	double&	operator [] (int i) {
			switch (i) {
				case 0:  return s;
				case 1:  return V.x;
				case 2:  return V.y;
				case 3:  return V.z;
				default:
					assert(0);
					return s;
			}
		}

	/** Set the components in Cartesian coordinate system. */
	void	set(const double as, const double x=0, const double y=0, const double z=0)
			{ s=as; V.set(x,y,z); }
	void	set(const double as, const Vector& v)
			{ s=as; V = v; }
	void	set(int i, double a) {
			switch (i) {
				case 0: s   = a; break;
				case 1: V.x = a; break;
				case 2: V.y = a; break;
				case 3: V.z = a; break;
				default: assert(0);
			}
		}

	/** @return square of quaternion q*q */
	Quaternion square()	const	{ return Quaternion(s*s - V.length2(), 2*s*V); }

	/** @return norm of quaternion */
	double	length2()	const	{ return s*s + V.length2(); }
	double	length()	const	{ return sqrt(length2()); }
	double	norm()		const	{ return length(); }

	/** @return a unit quaternion */
	Quaternion unit()	const	{ double n=norm(); return (*this)/n; }
	Quaternion versor()	const	{ return unit(); }

	/** normalize quaternion */
	double	normalize() {
			double n = norm();
			if (n > 0.0)
				*this /= n;
			return n;
		}

	/** return conjugate quaternion */
	Quaternion conjugate()	const	{ return Quaternion(s, -V); }
	Quaternion conj()	const	{ return conjugate(); }

	/** @return inverse quaternion */
	Quaternion inverse()	const	{ return conj() / length2(); }
	Quaternion inv()	const	{ return inverse(); }

	/** theta of unit quaternion.
	 * Unit quaternions can be written as (cos(Θ/2), sin(Θ/2) v) where |v|=1 */
	double	theta()	const	{ return 2*atan2(V.length(), s); }

	/** logarithm of unit quaternion */
	Quaternion log()	const	{ return Quaternion(0, theta()/2*V); }

	/** exponent of unit quaternion */
	Quaternion exp()	const {
			double th = theta()/2;
			return Quaternion(cos(th), sin(th)*V);
		}

	/** power of unit quaternion q^t = exp(t*log(q)) */
	Quaternion pow(double t) const	{ return (log()*t).exp(); }

	/** Quaternion Addition is commutative
	 *        (p+q)*r = pr + qr
	 *        p*(p+r) = pq + pr
	 * @param q quaternion to add * */
	Quaternion& operator += (const Quaternion& q) {
			s += q.s;
			V += q.V;
			return *this;
		}
	Quaternion operator + (const Quaternion& q) const { return Quaternion(s+q.s, V+q.V); }

	/** Subtraction
	 * @param q quaternion to add */
	Quaternion& operator -= (const Quaternion& q) {
			s -= q.s;
			V -= q.V;
			return *this;
		}
	Quaternion operator - (const Quaternion& q) const { return Quaternion(s-q.s, V-q.V); }

	/** Unary minus. */
	Quaternion operator - () const { return Quaternion(-s, -V); }

	/** Multiply */
	Quaternion& operator *= (const double d) {
			s *= d;
			V *= d;
			return *this;
		}

	Quaternion operator * (const double d) const { return Quaternion(d*s, d*V); }

	/** Multiply operators they are associative (pq)r = p(qr) */
	Quaternion operator * (const Quaternion& q) const
			{ return Quaternion(s*q.s-V*q.V,  s*q.V + q.s*V + (V^q.V)); }

	/** Multiply with a vector assume V.s = 0 */
	Quaternion operator * (const Vector& Q) const
			{ return Quaternion(-V*Q,  s*Q + (V^Q)); }

	/** Division operators */
	Quaternion& operator /= (const double d) {
			s /= d;
			V /= d;
			return *this;
		}

	Quaternion operator / (const double d) const { return Quaternion(s/d, V/d); }

	Quaternion operator / (const Quaternion& q) const
			{ return (*this) * q.inverse(); }

	/* --- Rotations --- */
	/** rotate vector with quaternion w = q v q* */
	Vector	rotate(const Vector v) const { return (((*this)*v)*conj()).V; }

	/** get rotation matrix */
	void matrix(Matrix4& M) const {
		assert(Eq(length(),1.0,1e-10));
//		//double cosq2_square = s*s;			// cos(θ/2) = s
//		double sinq2_square = V.length2();		// sin(θ/2) = V.length()
//		// Using double angle formulas
//		double sinq = 2*sqrt(sinq2_square)*s;		// sin2a = 2*sina*cosa
//		double cosq = (Sqrt2*s-1)*(Sqrt2*s+1);		// cos2a = cos²a - sin²a
								//       = 1 - 2*sin²a
								//       = 2*cos²a - 1 = (sqrt(2)*cosa + 1) * (sqrt(2)*cosa - 1)
		/* Rotation matrix: if v=(l,m,n)
		 *    /x'\   / l²(1-cosΘ)+ cosΘ     lm(1-cosΘ)-nsinΘ    nl(1-cosΘ)+msinΘ \ /x\
		 *    |y'| = | lm(1-cosΘ)+nsinΘ     m²(1-cosΘ)+ cosΘ    mn(1-cosΘ)-lsinΘ | |y|
		 *    \z'/   \ nl(1-cosΘ)-msinΘ     mn(1-cosΘ)+lcosΘ    n²(1-cosΘ)+ cosΘ / \z/
		 */
//		M.zero();
//		M(0,0) = Sq(V.x)*(1-cosq) +     cosq;
//		M(0,1) = V.x*V.y*(1-cosq) - V.z*sinq;
//		M(0,2) = V.z*V.x*(1-cosq) + V.y*sinq;
//
//		M(1,0) = V.x*V.y*(1-cosq) + V.z*sinq;
//		M(1,1) = Sq(V.y)*(1-cosq) +     cosq;
//		M(1,2) = V.y*V.z*(1-cosq) - V.x*sinq;
//
//		M(2,0) = V.z*V.x*(1-cosq) - V.y*sinq;
//		M(2,1) = V.y*V.z*(1-cosq) + V.x*cosq;
//		M(2,2) = Sq(V.z)*(1-cosq) +     cosq;

		M.zero();
		M(0,0) = 1-2*( Sq(V[1]) + Sq(V[2]));
		M(0,1) =   2*(V[0]*V[1] - V[2]*s);
		M(0,2) =   2*(V[2]*V[0] + V[1]*s);

		M(1,0) =   2*(V[0]*V[1] + V[2]*s);
		M(1,1) = 1-2*( Sq(V[2]) + Sq(V[0]));
		M(1,2) =   2*(V[1]*V[2] - V[0]*s);

		M(2,0) =   2*(V[2]*V[0] - V[1]*s);
		M(2,1) =   2*(V[1]*V[2] + V[0]*s);
		M(2,2) = 1-2*( Sq(V[1]) + Sq(V[0]));

		M(3,3) = 1;
	}
}; // class Quaternion

// Output to a stream.
std::ostream& operator << (std::ostream&, const Quaternion&);

// Scaling of quaternions with a real number
inline Quaternion operator * (const double a, const Quaternion& q) {
	return Quaternion(a*q.s, a*q.V);
}

/** logarithm of unit quaternion */
inline Quaternion log(const Quaternion& q) {
	return Quaternion(0, q.theta()/2*q.V);
}

/** exponent of unit quaternion */
inline Quaternion exp(const Quaternion& q) {
	double th = q.theta()/2;
	return Quaternion(cos(th), sin(th)*q.V);
}

/** multiply vector with quaternion */
inline Quaternion operator *(const Vector& V, const Quaternion& q) {
	return Quaternion(-V*q.V,  q.s*V + (V^q.V));
}

/** power of unit quaternion q^t = exp(t*log(q)) */
inline Quaternion pow(const Quaternion& q, const double t) {
	return exp(t*log(q));
}

#endif
