/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * FIXME:
 *  - propagate dose error to dvh
 *
 * Authors:	Vasilis.Vlachoudis@cern.ch
 *		Christian.Theis@cern.ch
 * Date:	02-Feb-2015
 */

#include <stdio.h>
#include <assert.h>
#include <getopt.h>

#include <mutex>
#include <future>
#include <thread>
#include <vector>
#include <utility>
#include <iostream>

#include "os.h"
#include "data/usrbin.h"
#include "math/random.h"
#include "math/histogram.h"
#include "geometry/voxel.h"

#ifdef MEM
#include "tool/memory.h"
#endif

using std::cout;
using std::cerr;
using std::endl;

/** global variables */
char*	prgname;
Usrbin	usrbin;
GVoxel	voxel;
bool	same;			// flag if usrbin has same dimensions as the voxel
bool	scan = true;		// Scan to find maximum value of each ROI
				// otherwise use the default max of the 
double	relative = 0.0;
double	norm     = 1.0;
Vector	U;
Vector	V;
Vector	W;
int	nbins    = 100;
int	nsamples = 5;		// QMC super sampling
std::vector<double> samples;	// sampling position vector

#ifdef THREAD
#	define _THREAD(x)	x
#else
#	define _THREAD(x)
#endif

// ==================================== DVH ====================================
class DVH {
public:
	int	id;
	int	volume;
	double	maxdose;
	H1D	histogram;
private:
	_THREAD(std::mutex	mutex;)
public:
	DVH() : id(0), volume(0), maxdose(0) {}
	DVH(const DVH& r) :			// we need this to use the vector.resize()
			id(r.id),		// otherwise it complains about the mutex
			volume(r.volume),	// not being copyable
			maxdose(r.maxdose),
			histogram(r.histogram)
			{}
	DVH& operator = (const DVH& r) {	// same as before
			id        = r.id;
			volume    = r.volume;
			maxdose   = r.maxdose;
			histogram = r.histogram;
			return *this;
		}

	/* compare maxdose */
	void	cmpDose(double d)	{ maxdose = Max(maxdose, d); }
	void	fill(const double v)	{
			_THREAD(lock());
			histogram.fill(v);
			_THREAD(unlock());
		}
#ifdef THREAD
	void	lock()		{ mutex.lock(); }
	void	unlock()	{ mutex.unlock(); }
#endif
}; // class DVH histogram information

std::vector<DVH>	dvh;

//int	*volumes=NULL;	// array with ROI structure volume information (should move to GVoxel)
//double	*maxdose=NULL;	// array with maximum dose information

/** check if usrbin and voxel have the same binning */
bool checkUsrbinNVoxel()
{
	const double eps = 1e-7;

	if (usrbin.nx!=voxel.nx) return false;
	if (usrbin.ny!=voxel.ny) return false;
	if (usrbin.nz!=voxel.nz) return false;

	if (Eq(usrbin.xlow, voxel.xlow, eps)) return false;
	if (Eq(usrbin.xlow, voxel.xlow, eps)) return false;
	if (Eq(usrbin.xlow, voxel.xlow, eps)) return false;

	if (Eq(usrbin.xhigh, voxel.xhigh, eps)) return false;
	if (Eq(usrbin.xhigh, voxel.xhigh, eps)) return false;
	if (Eq(usrbin.xhigh, voxel.xhigh, eps)) return false;

	return true;
} // checkUsrbinNVoxel

/** latinHypercube */
void latinHypercube(int n, int ndim, Random& random)
{
	samples.resize(nsamples*3);

	// Generate normalized LHS samples along diagonal
	double delta = 1.0 / n;

	for (int i=0; i<n; i++)
		for (int j=0; j<ndim; j++)
			samples[ndim*i + j] = (i + random.real()) * delta;

	// Permute LHS samples in each dimension
	for (int i=0; i<ndim; i++)
		for (int j=0; j<n; j++) {
			Random::Int other = random.integer(n);
			Swap(samples[ndim*j + i], samples[ndim * other + i]);
		}
} // latinHypercube

/** scanROI: scan voxel for structures return volumes and maximum dose per structure */
void scanROI()
{
	for (int k=0; k<voxel.nz; k++) {
		double z = voxel.voxelcz(k);
		for (int j=0; j<voxel.ny; j++) {
			double y = voxel.voxelcy(j);
			for (int i=0; i<voxel.nx; i++) {
				double x = voxel.voxelcx(i);
				const ROICombination& comb = voxel.roiComb(i,j,k);
				double dose;
				if (same)
					dose = norm * usrbin.get(i,j,k);
				else
				if (usrbin.get(x,y,z,&dose))
					dose *= norm;

				for (int m=0; m<comb.length; m++) {
					int roi = abs(comb[m]);
					dvh[roi].volume++;
					dvh[roi].cmpDose(dose);
				}
			}
		}
	}

	/* print maximum doses and values */
	printf("%3s %-24s %-10s %-10s\n", "ROI", "Name", "Vol", "Max Dose");
	double dxdydz = voxel.dx * voxel.dy * voxel.dz;
	for (int i=0; i<=voxel.nroi(); i++) {
		std::string name = i==0? "void" : voxel.roiName(i);
		if (name.empty()) continue;
		printf("%3d %-24s %10.5g %10.5g\n", i, name.c_str(),
				dxdydz*(double)dvh[i].volume, dvh[i].maxdose);
	}
} // scanROI

/** fillDVH */
void fillDVH()
{
	// prepare histograms
	if (scan) {
		dvh[0].histogram.title("void");
		dvh[0].histogram.set(nbins, 0.0, dvh[0].maxdose);
		for (int i=1; i<=voxel.nroi(); i++) {
			dvh[i].histogram.title(voxel.roiName(i));
			if (dvh[i].maxdose > 0.0)
				dvh[i].histogram.set(nbins, 0.0, dvh[i].maxdose);
			else
				dvh[i].histogram.set(nbins, 0.0, usrbin.max);
		}
	} else
		for (int i=1; i<=voxel.nroi(); i++) {
			if (voxel.roiName(i).empty()) continue;
			dvh[i].histogram.title(voxel.roiName(i));
			dvh[i].histogram.set(nbins, 0.0, usrbin.max);
		}

	if (not same) {
		if (nsamples==1) {
			samples.resize(nsamples*3);
			samples[0] = voxel.dx/2;
			samples[1] = voxel.dy/2;
			samples[2] = voxel.dz/2;
		} else {
			Random random(123);
			latinHypercube(nsamples, 3, random);
			// rescale with voxel dimensions
			for (int i=0; i<(int)samples.size();) {
				samples[i++] *= voxel.dx;
				samples[i++] *= voxel.dy;
				samples[i++] *= voxel.dz;
			}
		}
	}

	int perc = 10;
#ifndef THREAD
	// Fill the histograms
	for (int k=0; k<voxel.nz; k++) {
		double z = voxel.voxelz(k);
		for (int j=0; j<voxel.ny; j++) {
			double y = voxel.voxely(j);
			for (int i=0; i<voxel.nx; i++) {
				double x = voxel.voxelx(i);
				const ROICombination& comb = voxel.roiComb(i,j,k);
				double dose;
				if (same)
					dose = norm * usrbin.get(i,j,k);
				else {
					dose = 0.0;
					for (int n=0; n<(int)samples.size();) {
						double xn = x + samples[n++];
						double yn = y + samples[n++];
						double zn = z + samples[n++];
						//assert(InRange(x, xn, x+voxel.dx));
						//assert(InRange(y, yn, y+voxel.dy));
						//assert(InRange(z, zn, z+voxel.dz));
						double d;
						if (usrbin.get(xn,yn,zn,&d))
							dose += d;
					}
					dose *= norm / (double)(nsamples);
				}
				if (dose>0.0)
					for (int m=0; m<comb.length; m++)
						dvh[abs(comb[m])].fill(dose);
			}
		}
		int p = (k*100)/voxel.nz;
		if (p>=perc) {
			cout << "Completed " << p << "%" << endl;
			perc += 10;
		}
	}
#else
	// --------------------------------------------------------
	// threaded version treating one line of usrbin bins in X
	// --------------------------------------------------------
	auto fillHisto = [](int startj, int endj, int k) {
		double z = voxel.voxelz(k);
		for (int j=startj; j<endj; j++) {
			double y = voxel.voxely(j);
			for (int i=0; i<voxel.nx; i++) {
				double x = voxel.voxelx(i);
				const ROICombination& comb = voxel.roiComb(i,j,k);
				double dose;
				if (same)
					dose = norm * usrbin.get(i,j,k);
				else {
					dose = 0.0;
					for (int n=0; n<(int)samples.size();) {
						double xn = x + samples[n++];
						double yn = y + samples[n++];
						double zn = z + samples[n++];
						//assert(InRange(x, xn, x+voxel.dx));
						//assert(InRange(y, yn, y+voxel.dy));
						//assert(InRange(z, zn, z+voxel.dz));
						double d;
						if (usrbin.get(xn,yn,zn,&d))
							dose += d;
					}
					dose *= norm / (double)(nsamples);
				}
				if (dose>0.0)
					for (int m=0; m<comb.length; m++)
						dvh[abs(comb[m])].fill(dose);
			}
		}
	};

	//  set up threads
	unsigned nb_threads_hint = std::thread::hardware_concurrency();
	unsigned nb_threads = nb_threads_hint == 0 ? 16 : (nb_threads_hint);
	int seg = voxel.ny / nb_threads;

	//launch threads - parallel for
	auto fut_vec = std::vector<std::future<void>>();
	fut_vec.reserve(nb_threads);
	for (int k=0; k<voxel.nz; k++) {
		fut_vec.clear();
		// set up threads
		for (unsigned curThread=0; curThread < nb_threads; curThread++) {
			int startj = seg * curThread;
			fut_vec.emplace_back(std::async(std::launch::async,
							fillHisto,
							startj,
							Min(startj+seg, voxel.ny),
							k));
		}
		// sync and collect
		for (auto& th : fut_vec) th.get();

		int p = (k*100)/voxel.nz;
		if (p>=perc) {
			cout << "Completed " << p << "%" << endl;
			perc += 10;
		}
	}
#endif
	cout << "# Done 100%" << endl;

	// convert to cumulative
	for (int i=0; i<=voxel.nroi(); i++) {
		if (dvh[i].histogram.title().empty()) continue;
		if (scan) {
			if (dvh[i].maxdose==0.0) continue;
			if (relative!=0.0)
				dvh[i].histogram.setLimits(0., dvh[i].maxdose/relative*100.0);
				//dvh[i].histogram.setLimits(0., 100.0);
		}
		dvh[i].histogram.cumulative(true);
		dvh[i].histogram.norm(100.0/dvh[i].histogram.sumw());
	}
} // fillDVH

/** writeDVH */
void writeDVH(FILE *fout)
{
	int det = 1;
	if (scan) {
		/* First write volumes and maxdose */
		/* print maximum doses and values */
		double dxdydz = voxel.dx * voxel.dy * voxel.dz;

		fprintf(fout,"# Detector: %d volumes, maxdose\n",det++);
		for (int i=0; i<voxel.nroi()+1; i++)
			fprintf(fout,"%2d \"%s\" %.5g %.5g\n",
				i,
				dvh[i].histogram.ctitle(),
				dxdydz*(double)dvh[i].volume,
				dvh[i].maxdose);
		fprintf(fout,"\n\n");
	} 
	for (int i=0; i<voxel.nroi()+1; i++) {
		if (dvh[i].histogram.title().empty()) continue;
		if (scan and dvh[i].maxdose==0.0) continue;
		fprintf(fout,"# Detector: %d %s [%d]\n",
				det++, dvh[i].histogram.ctitle(),i);
		dvh[i].histogram.save(fout,4);
		if (i<voxel.nroi()) fprintf(fout,"\n\n");
	}
} // writeDVH

/* --- usage --- */
void usage()
{
	cout << "syntax: " << prgname <<" [options] <infile>\n";
	cout << "options:\n";
	cout << "\t-h?, --help\t\tThis help page\n";
	cout << "\t-d, --detector #\tDetector index number starting from 1\n";
	cout << "\t-S, --scan\t\tScan on first pass to find the max of each ROI histogram (default)\n";
	cout << "\t-M, --noscan\t\tDon't scan and use the maximum of usrbin for each ROI histogram\n";
	cout << "\t-n, --norm #\t\tNormalization factor\n";
	cout << "\t-N, --bins #\t\tnumber of bins per histogram (from 0..MaxDose)\n";
	cout << "\t-o, --output file\tA tab_lis file with the dvh histograms\n";
	cout << "\t-r, --relative #\tRelative to target dose (Default absolute)\n";
	cout << "\t-u, --usrbin #\t\tUsrbin file to load\n";
	cout << "\t-v, --voxel #\t\tVoxel file containing the ROI STRUCTures\n";
	cout << "\t-x, --xvoxel #\t\tX voxel origin (cm)\n";
	cout << "\t-y, --yvoxel #\t\tY voxel origin (cm)\n";
	cout << "\t-z, --zvoxel #\t\tZ voxel origin (cm)\n";
	cout << "\t--ux #\t\t\tU vector x component = IOP[0]\n";
	cout << "\t--uy #\t\t\tU vector y component = IOP[1]\n";
	cout << "\t--uz #\t\t\tU vector z component = IOP[2]\n";
	cout << "\t--vx #\t\t\tV vector x component = IOP[3]\n";
	cout << "\t--vy #\t\t\tV vector y component = IOP[4]\n";
	cout << "\t--vz #\t\t\tV vector z component = IOP[5]\n";
	cout << "\nauthor: Vasilis.Vlachoudis@cern.ch (2015)\n";
	exit(0);
} // usage

/** main */
int main(int ac, char *av[])
{
	prgname = av[0];
	std::string	usrbin_file, voxel_file;
	int	detector=0;
	char*	fileout=NULL;
	FILE*	fout=stdout;

	U = Vector::Xo;
	V = Vector::Yo;
	W = Vector::Zo;
	while (true) {
		int option_index = 0;
		static struct option long_options[] = {
			{"bins",    required_argument, 0, 'N'},
			{"detector",required_argument, 0, 'd'},
			{"help",    no_argument,       0, 'h'},
			{"norm",    required_argument, 0, 'n'},
			{"output",  required_argument, 0, 'o'},
			{"relative",required_argument, 0, 'r'},
			{"sample",  required_argument, 0, 's'},
			{"scan",    no_argument,       0, 'S'},
			{"noscan",  no_argument,       0, 'M'},
			{"usrbin",  required_argument, 0, 'u'},
			{"ux",      required_argument, 0,  0},
			{"uy",      required_argument, 0,  0},
			{"uz",      required_argument, 0,  0},
			{"voxel",   required_argument, 0, 'v'},
			{"vx",      required_argument, 0,  0},
			{"vy",      required_argument, 0,  0},
			{"vz",      required_argument, 0,  0},
			{"xvoxel",  required_argument, 0, 'x'},
			{"yvoxel",  required_argument, 0, 'y'},
			{"zvoxel",  required_argument, 0, 'z'},
			{0, 0, 0, 0}
		};

		int c = getopt_long(ac, av, "h?SMd:n:s:o:r:u:v:x:y:z:N:",
				long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 0:
				if (option_index<3)
					U[option_index] = atof(optarg);
				else
				if (option_index<6)
					V[option_index-3] = atof(optarg);
				break;

			case '?':
			case 'h':
				usage();
				break;

			case 'd':
				detector = atoi(optarg);
				break;

			case 'S':
				scan = true;
				break;

			case 'M':
				scan = false;
				break;

			case 'n':
				norm = atof(optarg);
				if (norm<=0.0) {
					cerr << "ERROR invalid normalization value " << norm << endl;
					return -1;
				}
				break;

			case 'N':
				nbins = atoi(optarg);
				if (nbins<10 or nbins>10000) {
					cerr << "ERROR invalid number of bins " << nbins << endl;
					return -1;
				}
				break;

			case 's':
				nsamples = atoi(optarg);
				if (nsamples<1 or nsamples>1000) {
					cerr << "ERROR invalid number of super sampling"
						<< nsamples << endl;
					return -1;
				}
				break;

			case 'o':
				fileout = strdup(optarg);
				break;

			case 'r':
				relative = atof(optarg);
				break;

			case 'u':
				usrbin_file = optarg;
				break;

			case 'v':
				voxel_file = optarg;
				break;

			case 'x':
				voxel.xlow = atof(optarg);
				break;

			case 'y':
				voxel.ylow = atof(optarg);
				break;

			case 'z':
				voxel.zlow = atof(optarg);
				break;

			default:
				cerr << "ERROR: ?? getopt returned character code " << c << endl;
				usage();
		}
	}
//	if (optind < ac) {
//		if ((fin=fopen(av[optind],"rb"))==NULL) {
//			fprintf(stderr,"Error opening input file %s\n",av[2]);
//			return 2;
//		}
//	}
	if (usrbin_file.length()==0) {
		cerr << "ERROR: usrbin file was not provided\n" << endl;
		return 1;
	}
	if (detector <= 0) {
		cerr << "ERROR: usrbin detector index was not provided\n" << endl;
		return 2;
	}
	if (voxel_file.length()==0) {
		cerr << "ERROR: voxel file was not provided\n" << endl;
		return 3;
	}

	cout << prgname << endl;
	cout << "Usrbin:   " << usrbin_file << endl;
	cout << "Detector: " << detector << endl;
	if (!usrbin.load(usrbin_file.c_str(), detector)) {
		cerr << "ERROR: loading usrbin detector " << usrbin_file << endl;
		return 4;
	}

	cout << "Voxel:  " << voxel_file << endl;
	if (!voxel.load(voxel_file.c_str())) {
		cerr << "ERROR: loading voxel file " << voxel_file << endl;
		return 5;
	}
	//cout << voxel << endl;

	if (voxel.nroi()==0) {
		cerr << "ERROR: voxel do not contain ROI STRUCT information" << endl;
		return 6;
	}

	cout << "Scan:   " << scan << endl;

	if (!Eq(U.length(),1,1e-7) or !Eq(V.length(),1.,1e-7)) {
		cerr << "ERROR: IOP vectors are not normalized to 1.0" << endl;
		cerr << "U = " << U << endl;
		cerr << "V = " << V << endl;
		return 7;
	}
	// check for orthogonalize
	W = U ^ V;			// calculate W

	voxel.calcLimits();
	same = checkUsrbinNVoxel();

	dvh.resize(voxel.nroi()+1, DVH());
	if (scan) scanROI();
	fillDVH();

	if (fileout) {
		fout = fopen(fileout,"w");
		if (fout==NULL) {
			cerr << "ERROR: opening output file " << optarg << endl;
			return 7;
		}
	}
	writeDVH(fout);

	if (fout != stdout) fclose(fout);
	if (fileout) free(fileout);
} // main
