/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:        19 Feb 2013
 */

#include <string>
#include <fstream>
#include <iostream>

#include <stdio.h>
#include <stdlib.h>

#include "os.h"
#include "nox.h"

#define _DRAW_BBOX false

using std::cout;
using std::cerr;
using std::endl;

const char* SEP = " \t";

/** NoxGeometry */
NoxGeometry::NoxGeometry(const char *filename)
	: kernel(geometry),
	  viewer(geometry,kernel),
	  dump(false)
{
	geometry.backgroundColor(0x707070);
	if (filename) load(filename);
} // NoxGeometry

/** parse matrix */
void NoxGeometry::parseMatrix(Matrix4& matrix)
{
	matrix.identity();
	for (int row=0; row<4; row++)
		for (int col=0; col<4; col++) {
			const char *token = strtok(NULL,SEP);
			if (!token) break;
			matrix(row,col) = atof(token);
		}
} // parseMatrix

/** parseBody */
GBody* NoxGeometry::parseBody()
{
	GBody* body;
	char* type = strtok(NULL,SEP);
	char* name = strtok(NULL,SEP);

	double what[30];
	memset(what, 0, sizeof(what));

	if (!strcmp(type,"VOXELS")) {
		if (!geometry.voxel.load(name)) {
			cerr << "ERROR Unable to load voxel " << name << endl;
		}
		geometry.voxel.xlow = atof(strtok(NULL,SEP));
		geometry.voxel.ylow = atof(strtok(NULL,SEP));
		geometry.voxel.zlow = atof(strtok(NULL,SEP));
		geometry.voxel.calcLimits();

		// create a voxel body
		body = geometry.addBody("VOXEL", "RPP");
		what[0] = geometry.voxel.xlow;
		what[1] = geometry.voxel.xhigh;
		what[2] = geometry.voxel.ylow;
		what[3] = geometry.voxel.yhigh;
		what[4] = geometry.voxel.zlow;
		what[5] = geometry.voxel.zhigh;

	} else {
		body = geometry.addBody(name, type);

		int i=0;
		const char* token;
		while ((token = strtok(NULL,SEP)))
			what[i++] = atof(token);
	}

	try {
		body->set(what);
	} catch (std::string err) {
		cerr << err << endl;
	}
	body->create();
	if (dump) cout << "Body: " << name << " " << type << endl;
	return body;
} // parseBody

/** parseRegionExpr */
bool NoxGeometry::parseRegionExpr(GRegion* region, std::string* line)
{
	char* tokptr = line? &(*line)[0] : NULL;

	const char* token;
	while ((token=strtok(tokptr,SEP))) {
		tokptr = NULL;
		if (token[0]==0 or token[0]==':') continue;
		GZone::Format format;
		if (!strcmp(token,"STD"))
			format = GZone::Format::PRODUCT;
		else
		if (!strcmp(token,"RPN"))
			format = GZone::Format::RPN;
		else {
			cerr << "Region " << region->name()
				<< " invalid format " << token << endl;
			return false;
		}
		geometry.addZone(region,format);
		while ((token=strtok(NULL,SEP))) {
			if (token[0]==':')
				break;
			else
			if (!geometry.add2exp(region, token))
				cerr << "Region " << region->name()
					<< " invalid token:" << token << endl;
		}
		if (token==NULL) break;
	}
	return true;
} // parseRegionExpr

/** parseRegion */
GRegion* NoxGeometry::parseRegion()
{
	GRegion* region = geometry.addRegion(strtok(NULL,SEP));
	if (dump) cout << "Region: " << region->name() << endl;
	region->type((GRegion::Type)atoi(strtok(NULL,SEP)));
	strtok(NULL,SEP);
	//int rotdefi = atoi(strtok(NULL,SEP));// FIXME rotdefi ignore for the moment!!!!!
	//region->matrix = atoi(strtok(NULL,SEP));	FIXME
	region_material_t regmat;
	regmat.region_id = region->id();
	regmat.color = atoi(strtok(NULL,SEP));
	regmat.alpha = atoi(strtok(NULL,SEP));
	region_materials.push_back(regmat);
#if _DRAW_BBOX
	region->show = BIT_BBOX;
#endif
	parseRegionExpr(region);
	return region;
} // parseRegion

/** parseVoxel */
void NoxGeometry::parseVoxel()
{
	const char* cmd = strtok(NULL,SEP);
	if (!strcmp(cmd,"color")) {
		int   i     = atoi(strtok(NULL,SEP));
		dword color = (dword)atoi(strtok(NULL,SEP));
		viewer.voxel().color(i-1, color);
	}
} // parseVoxel

/** parse */
bool NoxGeometry::parse(std::ifstream& input)
{
	std::string line;
	bool hasMatrix = false;
	Matrix4 matrix;			// current matrix
	std::string token;		// current token
	GRegion* region = NULL;	// current region

	while (std::getline(input, line)) {
		// BNV hack. calling the c_str is not necessary here
		// however with this we ensure that it will insert
		// a \0 character to be used later by strtok()
		if (line.empty() or line.c_str()[0]=='#') continue;

		// continuation lines
		if (line[0]==' ' or line[0]=='\t') {
			// continuation line
			if (token=="region" and region!=NULL) {
				parseRegionExpr(region, &line);
			} else {
				cerr << "Invalid continuation line: " << line << endl;
				//cerr << "token=" << token << endl;
				//cerr << "region=" << region << endl;
			}
			continue;
		}

		// BNV hack to get a pointer to the c-string using
		// c_str() will return a const object we cannot modify.
		// strtok() will replace by '\0' char every token sep
		token = strtok(&line[0],SEP);

		// initialize current values
		region = NULL;

		if (token == "rotdefi") {
			//int idx = atoi(strtok(NULL,SEP));
			//parseMatrix(matrix);
			//geometry.rotdefi(idx,matrix);
		} else
		if (token == "matrix") {
			parseMatrix(matrix);
			hasMatrix = true;
			if (dump) cout << "\tMatrix" << endl;
		} else
		if (token == "body") {
			auto body = parseBody();
			if (hasMatrix) {
				body->matrix(matrix);
				body->transform();
				hasMatrix = false;
			}
		} else
		if (token == "region") {
			region = parseRegion();
			if (hasMatrix) {
				region->matrix(matrix);
				hasMatrix = false;
			}
		} else
		if (token == "voxel")
			parseVoxel();
		else
		if (token == "viewmatrix") {
			parseMatrix(matrix);
			if (dump) cout << "View Matrix:" << endl << matrix << endl;
			viewer.matrix(matrix);
		} else
		if (token == "extends") {
			double x1 = atof(strtok(NULL,SEP));
			double y1 = atof(strtok(NULL,SEP));
			double x2 = atof(strtok(NULL,SEP));
			double y2 = atof(strtok(NULL,SEP));
			if (dump) cout	<< "Extends [" << x1 << ", " << y1
					<< "] - [" << x2 << ", " << y2 << "]" << endl;
			viewer.window(x1,y1,x2,y2);
		} else
		if (token == "origin") {
			double x = atof(strtok(NULL,SEP));
			double y = atof(strtok(NULL,SEP));
			double z = atof(strtok(NULL,SEP));
			if (dump) cout	<< "Origin [" << x << ", " << y
					<< ", " << z << "]" << endl;
			viewer.origin(x,y,z);
		} else
		if (token == "zoom") {
			double z = atof(strtok(NULL,SEP));
			if (dump) cout	<< "Zoom " << z << endl;
			viewer.view().zoom(z);
		} else {
			cerr << "Skipping unknown token " << token << endl;
			return false;
		}
	}
	kernel.derive();
	assignRegionMaterials();	// must be called after derive
	return true;
} // parse

/** assignRegionMaterials */
void NoxGeometry::assignRegionMaterials()
{
	// assign region materials
	for (auto regmat : region_materials) {
		VRegion* region = kernel.getRegion(regmat.region_id);
		if (region) {
			region->color(regmat.color);
			region->alpha(regmat.alpha);
		}
	}
} // assignRegionMaterials

/** load */
bool NoxGeometry::load(const char* filename)
{
	try {
		std::ifstream input(filename);
		parse(input);
		input.close();
	} catch (...) {
		cerr << "Error opening file " << filename << endl;
		return false;
	}
	geometry.derive();
	return true;
} // load

/** cleanup */
void NoxGeometry::cleanup()
{
	geometry.cleanup();
} // cleanup
